#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import collections
import logging
import math
import random
import warnings
from pathlib import Path

import dask
import numpy as np
from dask.base import tokenize
from dask.dataframe.core import new_dd_object
from dask.highlevelgraph import HighLevelGraph
from dask.utils import natural_sort_key, parse_bytes
from fsspec.core import get_fs_token_paths
from fsspec.utils import stringify_path

import merlin.core.dispatch as dispatch
from merlin.core.dispatch import convert_data, hex_to_int, is_dataframe_object
from merlin.core.utils import device_mem_size, global_dask_client, set_client_deprecated
from merlin.io.dataframe_iter import DataFrameIter
from merlin.io.shuffle import _check_shuffle_arg
from merlin.schema import ColumnSchema, Schema
from merlin.schema.io.tensorflow_metadata import TensorflowMetadata

from .csv import CSVDatasetEngine
from .dask import _ddf_to_dataset, _simple_shuffle
from .dataframe_engine import DataFrameDatasetEngine
from .parquet import ParquetDatasetEngine

try:
    import cudf
except ImportError:
    cudf = None

LOG = logging.getLogger("merlin")


class Dataset:
    """Universal external-data wrapper for NVTabular

    The NVTabular `Workflow` and `DataLoader`-related APIs require all
    external data to be converted to the universal `Dataset` type.  The
    main purpose of this class is to abstract away the raw format of the
    data, and to allow other NVTabular classes to reliably materialize a
    `dask_cudf.DataFrame` collection (and/or collection-based iterator)
    on demand.

    A new `Dataset` object can be initialized from a variety of different
    raw-data formats. To initialize an object from a directory path or
    file list, the `engine` argument should be used to specify either
    "parquet" or "csv" format.  If the first argument contains a list
    of files with a suffix of either "parquet" or "csv", the engine can
    be inferred::

        # Initialize Dataset with a parquet-dataset directory.
        # must specify engine="parquet"
        dataset = Dataset("/path/to/data_pq", engine="parquet")

        # Initialize Dataset with list of csv files.
        # engine="csv" argument is optional
        dataset = Dataset(["file_0.csv", "file_1.csv"])

    Since NVTabular leverages `fsspec` as a file-system interface,
    the underlying data can be stored either locally, or in a remote/cloud
    data store.  To read from remote storage, like gds or s3, the
    appropriate protocol should be prepended to the `Dataset` path
    argument(s), and any special backend parameters should be passed
    in a `storage_options` dictionary::

        # Initialize Dataset with s3 parquet data
        dataset = Dataset(
            "s3://bucket/path",
            engine="parquet",
            storage_options={'anon': True, 'use_ssl': False},
        )

    By default, both parquet and csv-based data will be converted to
    a Dask-DataFrame collection with a maximum partition size of
    roughly 12.5 percent of the total memory on a single device.  The
    partition size can be changed to a different fraction of total
    memory on a single device with the `part_mem_fraction` argument.
    Alternatively, a specific byte size can be specified with the
    `part_size` argument::

        # Dataset partitions will be ~10% single-GPU memory (or smaller)
        dataset = Dataset("bigfile.parquet", part_mem_fraction=0.1)

        # Dataset partitions will be ~1GB (or smaller)
        dataset = Dataset("bigfile.parquet", part_size="1GB")

    Note that, if both the fractional and literal options are used
    at the same time, `part_size` will take precedence.  Also, for
    parquet-formatted data, the partitioning is done at the row-
    group level, and the byte-size of the first row-group (after
    CuDF conversion) is used to map all other partitions.
    Therefore, if the distribution of row-group sizes is not
    uniform, the partition sizes will not be balanced.

    In addition to handling data stored on disk, a `Dataset` object
    can also be initialized from an existing CuDF/Pandas DataFrame,
    or from a Dask-DataFrame collection (e.g. `dask_cudf.DataFrame`).
    For these in-memory formats, the size/number of partitions will
    not be modified.  That is, a CuDF/Pandas DataFrame (or PyArrow
    Table) will produce a single-partition collection, while the
    number/size of a Dask-DataFrame collection will be preserved::

        # Initialize from CuDF DataFrame (creates 1 partition)
        gdf = cudf.DataFrame(...)
        dataset = Dataset(gdf)

        # Initialize from Dask-CuDF DataFrame (preserves partitions)
        ddf = dask_cudf.read_parquet(...)
        dataset = Dataset(ddf)

    Since the `Dataset` API can both ingest and output a Dask
    collection, it is straightforward to transform data either before
    or after an NVTabular workflow is executed. This means that some
    complex pre-processing operations, that are not yet supported
    in NVTabular, can still be accomplished with the Dask-CuDF API::

        # Sort input data before final Dataset initialization
        # Warning: Global sorting requires significant device memory!
        ddf = Dataset("/path/to/data_pq", engine="parquet").to_ddf()
        ddf = ddf.sort_values("user_rank", ignore_index=True)
        dataset = Dataset(ddf)

    `Dataset Optimization Tips (DOTs)`

    The NVTabular dataset should be created from Parquet files in order
    to get the best possible performance, preferably with a row group size
    of around 128MB.  While NVTabular also supports reading from CSV files,
    reading CSV can be over twice as slow as reading from Parquet. Take a
    look at this notebook_ for an example of transforming the original Criteo
    CSV dataset into a new Parquet dataset optimized for use with NVTabular.

    .. _notebook: https://github.com/NVIDIA/NVTabular/blob/main/examples/optimize_criteo.ipynb


    Parameters
    -----------
    path_or_source : str, list of str, or <dask.dataframe|cudf|pd>.DataFrame
        Dataset path (or list of paths), or a DataFrame. If string,
        should specify a specific file or directory path. If this is a
        directory path, the directory structure must be flat (nested
        directories are not yet supported).
    engine : str or DatasetEngine
        DatasetEngine object or string identifier of engine. Current
        string options include: ("parquet", "csv", "avro"). This argument
        is ignored if path_or_source is a DataFrame type.
    npartitions : int
        Desired number of Dask-collection partitions to produce in
        the ``to_ddf`` method when ``path_or_source`` corresponds to a
        DataFrame type.  This argument is ignored for file-based
        ``path_or_source`` input.
    part_size : str or int
        Desired size (in bytes) of each Dask partition.
        If None, part_mem_fraction will be used to calculate the
        partition size.  Note that the underlying engine may allow
        other custom kwargs to override this argument. This argument
        is ignored if path_or_source is a DataFrame type.
    part_mem_fraction : float (default 0.125)
        Fractional size of desired dask partitions (relative
        to GPU memory capacity). Ignored if part_size is passed
        directly. Note that the underlying engine may allow other
        custom kwargs to override this argument. This argument
        is ignored if path_or_source is a DataFrame type. If
        ``cpu=True``, this value will be relative to the total
        host memory detected by the client process.
    storage_options: None or dict
        Further parameters to pass to the bytes backend. This argument
        is ignored if path_or_source is a DataFrame type.
    cpu : bool
        WARNING: Experimental Feature!
        Whether NVTabular should keep all data in cpu memory when
        the Dataset is converted to an internal Dask collection. The
        default value is False, unless ``cudf`` and ``dask_cudf``
        are not installed (in which case the default is True). In the
        future, if True, NVTabular will NOT use any available GPU
        devices for down-stream processing.
        NOTE: Down-stream ops and output do not yet support a
        Dataset generated with ``cpu=True``.
    base_dataset : Dataset
        Optional reference to the original "base" Dataset object used
        to construct the current Dataset instance.  This object is
        used to preserve file-partition mapping information.
    schema : Schema
        Optional argument, to support custom user defined Schemas.
        This overrides the derived schema behavior.
    **kwargs :
        Key-word arguments to pass through to Dask.dataframe IO function.
        For the Parquet engine(s), notable arguments include `filters`,
        `aggregate_files`, and `gather_statistics`. Note that users who
        do not need to know the number of rows in their dataset (and do
        not plan to preserve a file-partition mapping) may wish to use
        `gather_statistics=False` for better client-side performance.
    """

    def __init__(
        self,
        path_or_source,
        engine=None,
        npartitions=None,
        part_size=None,
        part_mem_fraction=None,
        storage_options=None,
        dtypes=None,
        client="auto",
        cpu=None,
        base_dataset=None,
        schema=None,
        **kwargs,
    ):
        if schema is not None and not isinstance(schema, Schema):
            raise TypeError(f"unsupported schema type for merlin.io.Dataset: {type(schema)}")

        # Deprecate `client`
        if client != "auto":
            set_client_deprecated(client, "Dataset")

        self.dtypes = dtypes
        self.schema = schema

        # Cache for "real" (sampled) metadata
        self._real_meta = {}

        # Check if we are keeping data in cpu memory
        self.cpu = cpu
        if not self.cpu:
            self.cpu = cudf is None

        # Keep track of base dataset (optional)
        self.base_dataset = base_dataset or self

        # For now, lets warn the user that "cpu mode" is experimental
        if self.cpu:
            warnings.warn(
                "Initializing an NVTabular Dataset in CPU mode."
                "This is an experimental feature with extremely limited support!"
            )

        npartitions = npartitions or 1
        if isinstance(path_or_source, dask.dataframe.DataFrame) or is_dataframe_object(
            path_or_source
        ):
            # User is passing in a <dask.dataframe|cudf|pd>.DataFrame
            # Use DataFrameDatasetEngine
            _path_or_source = convert_data(
                path_or_source, cpu=self.cpu, to_collection=True, npartitions=npartitions
            )
            # Check if this is a collection that has now moved between host <-> device
            moved_collection = isinstance(path_or_source, dask.dataframe.DataFrame) and (
                not isinstance(_path_or_source._meta, type(path_or_source._meta))
            )
            if part_size:
                warnings.warn("part_size is ignored for DataFrame input.")
            if part_mem_fraction:
                warnings.warn("part_mem_fraction is ignored for DataFrame input.")
            self.engine = DataFrameDatasetEngine(
                _path_or_source, cpu=self.cpu, moved_collection=moved_collection
            )
        else:
            if part_size:
                # If a specific partition size is given, use it directly
                part_size = parse_bytes(part_size)
            else:
                # If a fractional partition size is given, calculate part_size
                part_mem_fraction = part_mem_fraction or 0.125
                assert 0.0 < part_mem_fraction < 1.0
                if part_mem_fraction > 0.25:
                    warnings.warn(
                        "Using very large partitions sizes for Dask. "
                        "Memory-related errors are likely."
                    )
                part_size = int(device_mem_size(kind="total", cpu=self.cpu) * part_mem_fraction)

            # Engine-agnostic path handling
            paths = stringify_path(path_or_source)
            if isinstance(paths, str):
                paths = [paths]
            paths = sorted(paths, key=natural_sort_key)

            storage_options = storage_options or {}
            # If engine is not provided, try to infer from end of paths[0]
            if engine is None:
                engine = paths[0].split(".")[-1]
            if isinstance(engine, str):
                if engine == "parquet":
                    self.engine = ParquetDatasetEngine(
                        paths, part_size, storage_options=storage_options, cpu=self.cpu, **kwargs
                    )
                elif engine == "csv":
                    self.engine = CSVDatasetEngine(
                        paths, part_size, storage_options=storage_options, cpu=self.cpu, **kwargs
                    )
                elif engine == "avro":
                    try:
                        from .avro import AvroDatasetEngine
                    except ImportError as e:
                        raise RuntimeError(
                            "Failed to import AvroDatasetEngine. Make sure uavro is installed."
                        ) from e

                    self.engine = AvroDatasetEngine(
                        paths, part_size, storage_options=storage_options, cpu=self.cpu, **kwargs
                    )
                else:
                    raise ValueError("Only parquet, csv, and avro supported (for now).")
            else:
                self.engine = engine(
                    paths, part_size, cpu=self.cpu, storage_options=storage_options
                )

        # load in schema or infer if not available
        # path is always a list at this point
        if not self.schema:
            if isinstance(path_or_source, (str, Path)):
                path_or_source = [Path(path_or_source)]
            if isinstance(path_or_source, list) and isinstance(path_or_source[0], (str, Path)):
                # list of paths to files
                schema_path = Path(path_or_source[0])
                if schema_path.is_file():
                    schema_path = schema_path.parent

                if (schema_path / "schema.pbtxt").exists():
                    schema = TensorflowMetadata.from_proto_text_file(schema_path)
                    self.schema = schema.to_merlin_schema()
                elif (schema_path.parent / "schema.pbtxt").exists():
                    schema = TensorflowMetadata.from_proto_text_file(schema_path.parent)
                    self.schema = schema.to_merlin_schema()
                else:
                    self.infer_schema()
            else:
                # df with no schema
                self.infer_schema()

    def to_ddf(self, columns=None, shuffle=False, seed=None):
        """Convert `Dataset` object to `dask_cudf.DataFrame`

        Parameters
        -----------
        columns : str or list(str); default None
            Columns to include in output `DataFrame`. If not specified,
            the output will contain all known columns in the Dataset.
        shuffle : bool; default False
            Whether to shuffle the order of partitions in the output
            `dask_cudf.DataFrame`.  Note that this does not shuffle
            the rows within each partition. This is because the data
            is not actually loaded into memory for this operation.
        seed : int; Optional
            The random seed to use if `shuffle=True`.  If nothing
            is specified, the current system time will be used by the
            `random` std library.
        """
        # Use DatasetEngine to create ddf
        ddf = self.engine.to_ddf(columns=columns)

        # Shuffle the partitions of ddf (optional)
        if shuffle and ddf.npartitions > 1:
            # Start with ordered partitions
            inds = list(range(ddf.npartitions))

            # Use random std library to reorder partitions
            random.seed(seed)
            random.shuffle(inds)

            # Construct new high-level graph (HLG)
            name = ddf._name
            new_name = "shuffle-partitions-" + tokenize(ddf)
            dsk = {(new_name, i): (lambda x: x, (name, ind)) for i, ind in enumerate(inds)}

            new_graph = HighLevelGraph.from_collections(new_name, dsk, dependencies=[ddf])

            # Convert the HLG to a Dask collection
            divisions = [None] * (ddf.npartitions + 1)
            ddf = new_dd_object(new_graph, new_name, ddf._meta, divisions)

        # Special dtype conversion (optional)
        if self.dtypes:
            _meta = _set_dtypes(ddf._meta, self.dtypes)
            return ddf.map_partitions(_set_dtypes, self.dtypes, meta=_meta)
        return ddf

    @property
    def file_partition_map(self):
        return self.engine._file_partition_map

    @property
    def partition_lens(self):
        return self.engine._partition_lens

    def to_cpu(self):
        warnings.warn(
            "Changing an NVTabular Dataset to CPU mode."
            "This is an experimental feature with extremely limited support!"
        )
        self.cpu = True
        self.engine.to_cpu()

    def to_gpu(self):
        self.cpu = False
        self.engine.to_gpu()

    def shuffle_by_keys(self, keys, hive_data=None, npartitions=None):
        """Shuffle the in-memory Dataset so that all unique-key
        combinations are moved to the same partition.

        Parameters
        ----------
        keys : list(str)
            Column names to shuffle by.
        hive_data : bool; default None
            Whether the dataset is backed by a hive-partitioned
            dataset (with the keys encoded in the directory structure).
            By default, the Dataset's `file_partition_map` property will
            be inspected to infer this setting. When `hive_data` is True,
            the number of output partitions will correspond to the number
            of unique key combinations in the dataset.
        npartitions : int; default None
            Number of partitions in the output Dataset. For hive-partitioned
            data, this value should be <= the number of unique key
            combinations (the default), otherwise it will be ignored. For
            data that is not hive-partitioned, the ``npartitions`` input
            should be <= the original partition count, otherwise it will be
            ignored.
        """

        # Make sure we are dealing with a list
        keys = [keys] if not isinstance(keys, (list, tuple)) else keys

        # Start with default ddf
        ddf = self.to_ddf()
        if npartitions:
            npartitions = min(ddf.npartitions, npartitions)

        if hive_data is not False:
            # The keys may be encoded in the directory names.
            # Let's use the file_partition_map to extract this info.
            try:
                _mapping = self.file_partition_map
            except AttributeError as e:
                _mapping = None
                if hive_data:
                    raise RuntimeError("Failed to extract hive-partition mapping!") from e

            # If we have a `_mapping` available, check if the
            # file names include information about all our keys
            hive_mapping = collections.defaultdict(list)
            if _mapping:
                for k, v in _mapping.items():
                    for part in k.split(self.engine.fs.sep)[:-1]:
                        try:
                            _key, _val = part.split("=")
                        except ValueError:
                            continue
                        if _key in keys:
                            hive_mapping[_key].append(_val)

            if set(hive_mapping.keys()) == set(keys):

                # Generate hive-mapping DataFrame summary
                hive_mapping = type(ddf._meta)(hive_mapping)
                cols = list(hive_mapping.columns)
                for c in keys:
                    typ = ddf._meta[c].dtype
                    if c in cols:
                        if typ == "category":
                            # Cannot cast directly to categorical unless we
                            # first cast to the underlying dtype of the categories
                            hive_mapping[c] = hive_mapping[c].astype(typ.categories.dtype)
                        hive_mapping[c] = hive_mapping[c].astype(typ)

                # Generate simple-shuffle plan
                target_mapping = hive_mapping.drop_duplicates().reset_index(drop=True)
                target_mapping.index.name = "_partition"
                hive_mapping.index.name = "_sort"
                target_mapping.reset_index(drop=False, inplace=True)
                plan = (
                    hive_mapping.reset_index()
                    .merge(target_mapping, on=cols, how="left")
                    .sort_values("_sort")["_partition"]
                )

                if hasattr(plan, "to_pandas"):
                    plan = plan.to_pandas()

                # Deal with repartitioning
                if npartitions and npartitions < len(target_mapping):
                    q = np.linspace(0.0, 1.0, num=npartitions + 1)
                    divs = plan.quantile(q)
                    partitions = divs.searchsorted(plan, side="right") - 1
                    partitions[(plan >= divs.iloc[-1]).values] = len(divs) - 2
                    plan = partitions.tolist()
                elif len(plan) != len(plan.unique()):
                    plan = plan.to_list()
                else:
                    # Plan is a unique 1:1 ddf partition mapping.
                    # We already have shuffled data.
                    return self

                # TODO: We should avoid shuffling the original ddf and
                # instead construct a new (more-efficent) graph to read
                # multiple files from each partition directory at once.
                # Generally speaking, we can optimize this code path
                # much further.
                return Dataset(_simple_shuffle(ddf, plan))

        # Fall back to dask.dataframe algorithm
        return Dataset(ddf.shuffle(keys, npartitions=npartitions))

    def repartition(self, npartitions=None, partition_size=None):
        """Repartition the underlying ddf, and return a new Dataset

        Parameters
        ----------
        npartitions : int; default None
            Number of partitions in output ``Dataset``. Only used if
            ``partition_size`` isn’t specified.
        partition_size : int or str; default None
            Max number of bytes of memory for each partition. Use
            numbers or strings like '5MB'. If specified, ``npartitions``
            will be ignored.
        """
        return Dataset(
            self.to_ddf()
            .clear_divisions()
            .repartition(
                npartitions=npartitions,
                partition_size=partition_size,
            )
        )

    @classmethod
    def merge(cls, left, right, **kwargs):
        """Merge two Dataset objects

        Produces a new Dataset object. If the ``cpu`` Dataset attributes
        do not match, the right side will be modified. See Dask-Dataframe
        ``merge`` documentation for more information. Example usage::

            ds_1 = Dataset("file.parquet")
            ds_2 = Dataset(cudf.DataFrame(...))
            ds_merged = Dataset.merge(ds_1, ds_2, on="foo", how="inner")

        Parameters
        ----------
        left : Dataset
            Left-side Dataset object.
        right : Dataset
            Right-side Dataset object.
        **kwargs :
            Key-word arguments to be passed through to Dask-Dataframe.
        """

        # Ensure both Dataset objects are either cudf or pandas based
        if left.cpu and not right.cpu:
            _right = cls(right.to_ddf())
            _right.to_cpu()
        elif not left.cpu and right.cpu:
            _right = cls(right.to_ddf())
            _right.to_gpu()
        elif left.cpu == right.cpu:
            # both left and right are already cudf / pandas df
            _right = right

        return cls(
            left.to_ddf()
            .clear_divisions()
            .merge(
                _right.to_ddf().clear_divisions(),
                **kwargs,
            )
        )

    def to_iter(
        self, columns=None, indices=None, shuffle=False, seed=None, use_file_metadata=None, epochs=1
    ):
        """Convert `Dataset` object to a `cudf.DataFrame` iterator.

        Note that this method will use `to_ddf` to produce a
        `dask_cudf.DataFrame`, and materialize a single partition for
        each iteration.

        Parameters
        ----------
        columns : str or list(str); default None
            Columns to include in each `DataFrame`. If not specified,
            the outputs will contain all known columns in the Dataset.
        indices : list(int); default None
            A specific list of partition indices to iterate over. If
            nothing is specified, all partitions will be returned in
            order (or the shuffled order, if `shuffle=True`).
        shuffle : bool; default False
            Whether to shuffle the order of `dask_cudf.DataFrame`
            partitions used by the iterator.  If the `indices`
            argument is specified, those indices correspond to the
            partition indices AFTER the shuffle operation.
        seed : int; Optional
            The random seed to use if `shuffle=True`.  If nothing
            is specified, the current system time will be used by the
            `random` std library.
        use_file_metadata : bool; Optional
            Whether to allow the returned ``DataFrameIter`` object to
            use file metadata from the ``base_dataset`` to estimate
            the row-count. By default, the file-metadata
            optimization will only be used if the current Dataset is
            backed by a file-based engine. Otherwise, it is possible
            that an intermediate transform has modified the row-count.
        epochs : int
            Number of dataset passes to include within a single iterator.
            This option is used for multi-epoch data-loading. Default is 1.
        """
        if isinstance(columns, str):
            columns = [columns]

        # Try to extract the row-size metadata
        # if we are not shuffling
        partition_lens_meta = None
        if not shuffle and use_file_metadata is not False:
            # We are allowed to use file metadata to calculate
            # partition sizes.  If `use_file_metadata` is None,
            # we only use metadata if `self` is backed by a
            # file-based engine (like "parquet").  Otherwise,
            # we cannot be "sure" that the metadata row-count
            # is correct.
            try:
                if use_file_metadata:
                    partition_lens_meta = self.base_dataset.partition_lens
                else:
                    partition_lens_meta = self.partition_lens
            except AttributeError:
                pass

        return DataFrameIter(
            self.to_ddf(columns=columns, shuffle=shuffle, seed=seed),
            indices=indices,
            partition_lens=partition_lens_meta,
            epochs=epochs,
        )

    def to_parquet(
        self,
        output_path,
        shuffle=None,
        preserve_files=False,
        output_files=None,
        out_files_per_proc=None,
        num_threads=0,
        dtypes=None,
        cats=None,
        conts=None,
        labels=None,
        suffix=".parquet",
        partition_on=None,
        method="subgraph",
        write_hugectr_keyset=False,
    ):
        """Writes out to a parquet dataset

        Parameters
        ----------
        output_path : string
            Path to write processed/shuffled output data
        shuffle : merlin.io.Shuffle enum
            How to shuffle the output dataset. For all options,
            other than `None` (which means no shuffling), the partitions
            of the underlying dataset/ddf will be randomly ordered. If
            `PER_PARTITION` is specified, each worker/process will also
            shuffle the rows within each partition before splitting and
            appending the data to a number (`out_files_per_proc`) of output
            files. Output files are distinctly mapped to each worker process.
            If `PER_WORKER` is specified, each worker will follow the same
            procedure as `PER_PARTITION`, but will re-shuffle each file after
            all data is persisted.  This results in a full shuffle of the
            data processed by each worker.  To improve performance, this option
            currently uses host-memory `BytesIO` objects for the intermediate
            persist stage. The `FULL` option is not yet implemented.
        partition_on : str or list(str)
            Columns to use for hive-partitioning.  If this option is used,
            `preserve_files`, `output_files`, and `out_files_per_proc`
            cannot be specified, and `method` will be ignored.  Also, the
            `PER_WORKER` shuffle will not be supported.
        preserve_files : bool
            Whether to preserve the original file-to-partition mapping of
            the base dataset. This option requires `method="subgraph"`, and is
            only available if the base dataset is known, and if it corresponds
            to csv or parquet format. If True, the `out_files_per_proc` option
            will be ignored. Default is False.
        output_files : dict, list or int
            The total number of desired output files. This option requires
            `method="subgraph"`, and the default value will be the number of Dask
            workers, multiplied by `out_files_per_proc`. For further output-file
            control, this argument may also be used to pass a dictionary mapping
            the output file names to partition indices, or a list of desired
            output-file names.
        out_files_per_proc : integer
            Number of output files that each process will use to shuffle an input
            partition. Default is 1. If `method="worker"`, the total number of output
            files will always be the total number of Dask workers, multiplied by this
            argument. If `method="subgraph"`, the total number of files is determined
            by `output_files` (and `out_files_per_proc` must be 1 if a dictionary is
            specified).
        num_threads : integer
            Number of IO threads to use for writing the output dataset.
            For `0` (default), no dedicated IO threads will be used.
        dtypes : dict
            Dictionary containing desired datatypes for output columns.
            Keys are column names, values are datatypes.
        suffix : str or False
            File-name extension to use for all output files. This argument
            is ignored if a specific list of file names is specified using
            the ``output_files`` option. If ``preserve_files=True``, this
            suffix will be appended to the original name of each file,
            unless the original extension is ".csv", ".parquet", ".avro",
            or ".orc" (in which case the old extension will be replaced).
        cats : list of str, optional
            List of categorical columns
        conts : list of str, optional
            List of continuous columns
        labels : list of str, optional
            List of label columns
        method : {"subgraph", "worker"}
            General algorithm to use for the parallel graph execution. In order
            to minimize memory pressure, `to_parquet` will use a `"subgraph"` by
            default. This means that we segment the full Dask task graph into a
            distinct subgraph for each output file (or output-file group). Then,
            each of these subgraphs is executed, in full, by the same worker (as
            a single large task). In some cases, it may be more ideal to prioritize
            concurrency. In that case, a worker-based approach can be used by
            specifying `method="worker"`.
        write_hugectr_keyset : bool, optional
            Whether to write a HugeCTR keyset output file ("_hugectr.keyset").
            Writing this file can be very slow, and should only be done if you
            are planning to ingest the output data with HugeCTR. Default is False.
        """

        if partition_on:

            # Check that the user is not expecting a specific output-file
            # count/structure that is not supported
            if output_files:
                raise ValueError("`output_files` not supported when `partition_on` is used.")
            if out_files_per_proc:
                raise ValueError("`out_files_per_proc` not supported when `partition_on` is used.")
            if preserve_files:
                raise ValueError("`preserve_files` not supported when `partition_on` is used.")

        else:

            # Check that method (algorithm) is valid
            if method not in ("subgraph", "worker"):
                raise ValueError(f"{method} not a recognized method for `Dataset.to_parquet`")

            # Deal with method-specific defaults
            if method == "worker":
                if output_files or preserve_files:
                    raise ValueError("output_files and preserve_files require `method='subgraph'`")
                output_files = False
            elif preserve_files and output_files:
                raise ValueError("Cannot specify both preserve_files and output_files.")
            elif not (output_files or preserve_files):
                # Default "subgraph" behavior - Set output_files to the
                # total umber of workers, multiplied by out_files_per_proc
                try:
                    nworkers = len(global_dask_client().cluster.workers)
                except AttributeError:
                    nworkers = 1
                output_files = nworkers * (out_files_per_proc or 1)

        # Replace None/False suffix argument with ""
        suffix = suffix or ""

        # Check shuffle argument
        shuffle = _check_shuffle_arg(shuffle)

        if isinstance(output_files, dict) or (not output_files and preserve_files):
            # Do not shuffle partitions if we are preserving files or
            # if a specific file-partition mapping is already specified
            ddf = self.to_ddf()
        else:
            ddf = self.to_ddf(shuffle=shuffle)

        # Deal with `method=="subgraph"`.
        # Convert `output_files` argument to a dict mapping
        if output_files:

            #   NOTES on `output_files`:
            #
            # - If a list of file names is specified, a contiguous range of
            #   output partitions will be mapped to each file. The same
            #   procedure is used if an integer is specified, but the file
            #   names will be written as "part_*".
            #
            # - When `output_files` is used, the `output_files_per_proc`
            #   argument will be interpreted as the desired number of output
            #   files to write within the same task at run time (enabling
            #   input partitions to be shuffled into multiple output files).
            #
            # - Passing a list or integer to `output_files` will preserve
            #   the original ordering of the input data as long as
            #   `out_files_per_proc` is set to `1` (or `None`), and
            #   `shuffle==None`.
            #
            # - If a dictionary is specified, excluded partition indices
            #   will not be written to disk.
            #
            # - To map multiple output files to a range of input partitions,
            #   dictionary-input keys should correspond to a tuple of file
            #   names.

            # Use out_files_per_proc to calculate how
            # many output files should be written within the
            # same subgraph.  Note that we must a
            files_per_task = out_files_per_proc or 1
            required_npartitions = ddf.npartitions
            if isinstance(output_files, int):
                required_npartitions = output_files
                files_per_task = min(files_per_task, output_files)
            elif isinstance(output_files, list):
                required_npartitions = len(output_files)
                files_per_task = min(files_per_task, len(output_files))
            elif out_files_per_proc:
                raise ValueError(
                    "Cannot specify out_files_per_proc if output_files is "
                    "defined as a dictionary mapping. Please define each "
                    "key in output_files as a tuple of file names if you "
                    "wish to have those files written by the same process."
                )

            # Repartition ddf if necessary
            if ddf.npartitions < required_npartitions:
                ddf = ddf.clear_divisions().repartition(npartitions=required_npartitions)

            # Construct an output_files dictionary if necessary
            if isinstance(output_files, int):
                output_files = [f"part_{i}" + suffix for i in range(output_files)]
            if isinstance(output_files, list):
                new = {}
                split = math.ceil(ddf.npartitions / len(output_files))
                for i in range(0, len(output_files), files_per_task):
                    fns = output_files[i : i + files_per_task]
                    start = i * split
                    stop = min(start + split * len(fns), ddf.npartitions)
                    if start < stop:
                        new[tuple(fns)] = np.arange(start, stop)
                # let user know they will not have expected number of output files.
                if len(new.keys()) < len(output_files):
                    warnings.warn(
                        f"Only created {len(new.keys())} files did not have enough "
                        f"partitions to create {len(output_files)} files."
                    )
                output_files = new
                suffix = ""  # Don't add a suffix later - Names already include it
            if not isinstance(output_files, dict):
                raise TypeError(f"{type(output_files)} not a supported type for `output_files`.")

        # If we are preserving files, use the stored dictionary,
        # or use file_partition_map to extract the mapping
        elif preserve_files:
            try:
                _output_files = self.base_dataset.file_partition_map
            except AttributeError as e:
                raise AttributeError(
                    f"`to_parquet(..., preserve_files=True)` is not currently supported "
                    f"for datasets with a {type(self.base_dataset.engine)} engine. Check "
                    f"that `dataset.base_dataset` is backed by csv or parquet files."
                ) from e
            if suffix == "":
                output_files = _output_files
            else:
                output_files = {}
                for fn, rgs in _output_files.items():
                    split_fn = fn.split(".")
                    if split_fn[-1] in ("parquet", "avro", "orc", "csv"):
                        output_files[".".join(split_fn[:-1]) + suffix] = rgs
                    else:
                        output_files[fn + suffix] = rgs
            suffix = ""  # Don't add a suffix later - Names already include it

        if dtypes:
            _meta = _set_dtypes(ddf._meta, dtypes)
            ddf = ddf.map_partitions(_set_dtypes, dtypes, meta=_meta)

        fs = get_fs_token_paths(output_path)[0]
        fs.mkdirs(output_path, exist_ok=True)

        tf_metadata = TensorflowMetadata.from_merlin_schema(self.schema)
        tf_metadata.to_proto_text_file(output_path)

        # Output dask_cudf DataFrame to dataset
        _ddf_to_dataset(
            ddf,
            fs,
            output_path,
            shuffle,
            output_files,
            out_files_per_proc,
            cats or [],
            conts or [],
            labels or [],
            "parquet",
            num_threads,
            self.cpu,
            suffix=suffix,
            partition_on=partition_on,
            schema=self.schema if write_hugectr_keyset else None,
        )

    def to_hugectr(
        self,
        output_path,
        cats,
        conts,
        labels,
        shuffle=None,
        file_partition_map=None,
        out_files_per_proc=None,
        num_threads=0,
        dtypes=None,
    ):
        """Writes out to a hugectr dataset

        Parameters
        ----------
        output_path : string
            Path to write processed/shuffled output data
        cats : list of str
            List of categorical columns
        conts : list of str
            List of continuous columns
        labels : list of str
            List of label columns
        shuffle : merlin.io.Shuffle, optional
            How to shuffle the output dataset. Shuffling is only
            performed if the data is written to disk. For all options,
            other than `None` (which means no shuffling), the partitions
            of the underlying dataset/ddf will be randomly ordered. If
            `PER_PARTITION` is specified, each worker/process will also
            shuffle the rows within each partition before splitting and
            appending the data to a number (`out_files_per_proc`) of output
            files. Output files are distinctly mapped to each worker process.
            If `PER_WORKER` is specified, each worker will follow the same
            procedure as `PER_PARTITION`, but will re-shuffle each file after
            all data is persisted.  This results in a full shuffle of the
            data processed by each worker.  To improve performance, this option
            currently uses host-memory `BytesIO` objects for the intermediate
            persist stage. The `FULL` option is not yet implemented.
        file_partition_map : dict
            Dictionary mapping of output file names to partition indices
            that should be written to that file name.  If this argument
            is passed, only the partitions included in the dictionary
            will be written to disk, and the `output_files_per_proc` argument
            will be ignored.
        out_files_per_proc : integer
            Number of files to create (per process) after
            shuffling the data
        num_threads : integer
            Number of IO threads to use for writing the output dataset.
            For `0` (default), no dedicated IO threads will be used.
        dtypes : dict
            Dictionary containing desired datatypes for output columns.
            Keys are column names, values are datatypes.
        """

        # For now, we must move to the GPU to
        # write an output dataset.
        # TODO: Support CPU-mode output
        self.to_gpu()

        shuffle = _check_shuffle_arg(shuffle)
        ddf = self.to_ddf(shuffle=shuffle)
        if dtypes:
            _meta = _set_dtypes(ddf._meta, dtypes)
            ddf = ddf.map_partitions(_set_dtypes, dtypes, meta=_meta)

        fs = get_fs_token_paths(output_path)[0]
        fs.mkdirs(output_path, exist_ok=True)
        self.schema.write(output_path)

        # Output dask_cudf DataFrame to dataset,
        _ddf_to_dataset(
            ddf,
            fs,
            output_path,
            shuffle,
            file_partition_map,
            out_files_per_proc,
            cats,
            conts,
            labels,
            "hugectr",
            num_threads,
            self.cpu,
            schema=self.schema,
        )

    @property
    def num_rows(self):
        return self.engine.num_rows

    @property
    def npartitions(self):
        return self.to_ddf().npartitions

    def validate_dataset(self, **kwargs):
        """Validate for efficient processing.

        The purpose of this method is to validate that the Dataset object
        meets the minimal requirements for efficient NVTabular processing.
        For now, this criteria requires the data to be in parquet format.

        Example Usage::

            dataset = Dataset("/path/to/data_pq", engine="parquet")
            assert validate_dataset(dataset)

        Parameters
        -----------
        **kwargs :
            Key-word arguments to pass down to the engine's validate_dataset
            method. For the recommended parquet format, these arguments
            include `add_metadata_file`, `row_group_max_size`, `file_min_size`,
            and `require_metadata_file`. For more information, see
            `ParquetDatasetEngine.validate_dataset`.

        Returns
        -------
        valid : bool
            `True` if the input dataset is valid for efficient NVTabular
            processing.
        """

        # Check that the dataset format is Parquet
        if not isinstance(self.engine, ParquetDatasetEngine):
            msg = (
                "NVTabular is optimized for the parquet format. Please use "
                "the to_parquet method to convert your dataset."
            )
            warnings.warn(msg)
            return False  # Early return

        return self.engine.validate_dataset(**kwargs)

    def regenerate_dataset(
        self,
        output_path,
        columns=None,
        output_format="parquet",
        compute=True,
        **kwargs,
    ):
        """EXPERIMENTAL:
        Regenerate an NVTabular Dataset for efficient processing by writing
        out new Parquet files. In contrast to default ``to_parquet`` behavior,
        this method preserves the original ordering.

        Example Usage::

            dataset = Dataset("/path/to/data_pq", engine="parquet")
            dataset.regenerate_dataset(
                out_path, part_size="1MiB", file_size="10MiB"
            )

        Parameters
        -----------
        output_path : string
            Root directory path to use for the new (regenerated) dataset.
        columns : list(string), optional
            Subset of columns to include in the regenerated dataset.
        output_format : string, optional
            Format to use for regenerated dataset.  Only "parquet" (default)
            is currently supported.
        compute : bool, optional
            Whether to compute the task graph or to return a Delayed object.
            By default, the graph will be executed.
        **kwargs :
            Key-word arguments to pass down to the engine's regenerate_dataset
            method. See `ParquetDatasetEngine.regenerate_dataset` for more
            information.

        Returns
        -------
        result : int or Delayed
            If `compute=True` (default), the return value will be an integer
            corresponding to the number of generated data files.  If `False`,
            the returned value will be a `Delayed` object.
        """

        # Check that the desired output format is Parquet
        if output_format not in ["parquet"]:
            msg = (
                f"NVTabular is optimized for the parquet format. "
                f"{output_format} is not yet a supported output format for "
                f"regenerate_dataset."
            )
            raise ValueError(msg)

        result = ParquetDatasetEngine.regenerate_dataset(self, output_path, columns=None, **kwargs)
        if compute:
            return result.compute()
        else:
            return result

    def infer_schema(self, n=1):
        """Create a schema containing the column names and inferred dtypes of the Dataset

        Args:
            n (int, optional): Number of rows to sample to infer the dtypes. Defaults to 1.
        """

        dtypes = {}

        dtypes = self.sample_dtypes(n=n, annotate_lists=True)

        column_schemas = []
        for column, dtype_info in dtypes.items():
            dtype_val = dtype_info["dtype"]
            is_list = dtype_info["is_list"]
            col_schema = ColumnSchema(column, dtype=dtype_val, is_list=is_list, is_ragged=is_list)
            column_schemas.append(col_schema)

        self.schema = Schema(column_schemas)
        return self.schema

    def sample_dtypes(self, n=1, annotate_lists=False):
        """Return the real dtypes of the Dataset

        Use cached metadata if this operation was
        already performed. Otherwise, call down to the
        underlying engine for sampling logic.
        """
        if self._real_meta.get(n, None) is None:
            _real_meta = self.engine.sample_data(n=n)
            if self.dtypes:
                _real_meta = _set_dtypes(_real_meta, self.dtypes)
            self._real_meta[n] = _real_meta

        if annotate_lists:
            _real_meta = self._real_meta[n]
            annotated = {
                col: {
                    "dtype": dispatch.list_val_dtype(_real_meta[col]) or _real_meta[col].dtype,
                    "is_list": dispatch.is_list_dtype(_real_meta[col]),
                }
                for col in _real_meta.columns
            }
            return annotated

        return self._real_meta[n].dtypes

    @classmethod
    def _bind_dd_method(cls, name):
        """Bind Dask-Dataframe method to the Dataset class"""

        def meth(self, *args, **kwargs):
            _meth = getattr(self.to_ddf(), name)
            return _meth(*args, **kwargs)

        meth.__name__ = name
        setattr(cls, name, meth)


# Bind (simple) Dask-Dataframe Methods
for op in ["compute", "persist", "head", "tail"]:
    Dataset._bind_dd_method(op)


def _set_dtypes(chunk, dtypes):
    for col, dtype in dtypes.items():
        if isinstance(dtype, str) and ("hex" in dtype):
            chunk[col] = hex_to_int(chunk[col])
        else:
            chunk[col] = chunk[col].astype(dtype)
    return chunk



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import contextlib
import threading

import pandas as pd

try:
    import cudf
except ImportError:
    cudf = None
import pyarrow as pa
from dask.distributed import get_worker

# Use global variable as the default
# cache when there are no distributed workers.
# Use a thread lock to "handle" multiple Dask
# worker threads.
_WORKER_CACHE = {}
_WORKER_CACHE_LOCK = threading.RLock()


@contextlib.contextmanager
def get_worker_cache(name):
    with _WORKER_CACHE_LOCK:
        yield _get_worker_cache(name)


def _get_worker_cache(name):
    """Utility to get the `name` element of the cache
    dictionary for the current worker.  If executed
    by anything other than a distributed Dask worker,
    we will use the global `_WORKER_CACHE` variable.
    """
    try:
        worker = get_worker()
    except ValueError:
        # There is no dask.distributed worker.
        # Assume client/worker are same process
        global _WORKER_CACHE  # pylint: disable=global-statement
        if name not in _WORKER_CACHE:
            _WORKER_CACHE[name] = {}
        return _WORKER_CACHE[name]
    if not hasattr(worker, "worker_cache"):
        worker.worker_cache = {}
    if name not in worker.worker_cache:
        worker.worker_cache[name] = {}
    return worker.worker_cache[name]


def fetch_table_data(
    table_cache, path, cache="disk", cats_only=False, reader=None, columns=None, **kwargs
):
    """Utility to retrieve a cudf DataFrame from a cache (and add the
    DataFrame to a cache if the element is missing).  Note that `cats_only=True`
    results in optimized logic for the `Categorify` transformation.
    """
    _lib = cudf if cudf else pd
    reader = reader or _lib.read_parquet
    table = table_cache.get(path, None)
    cache_df = cache == "device"
    if table is None:
        use_kwargs = {"columns": columns} if columns is not None else {}
        use_kwargs.update(kwargs)
        if cache in ("device", "disk"):
            table = reader(path, **use_kwargs)
        elif cache == "host":
            if reader == _lib.read_parquet:  # pylint: disable=comparison-with-callable
                # Using cudf-backed data with "host" caching.
                # Cache as an Arrow table.
                table = reader(path, **use_kwargs)
                if cudf:
                    table_cache[path] = table.to_arrow()
                else:
                    table_cache[path] = pa.Table.from_pandas(table)
                if columns is not None:
                    table = table[columns]
            else:
                # Using pandas-backed data with "host" caching.
                # Just read in data and cache as a pandas DataFrame.
                table = reader(path, **use_kwargs)
                cache_df = True
        if cats_only:
            table.index.name = "labels"
            table.reset_index(drop=False, inplace=True)
        if cache_df:
            table_cache[path] = table.copy(deep=False)
    elif isinstance(table, pa.Table):
        if cudf:
            df = cudf.DataFrame.from_arrow(table)
        else:
            df = table.to_pandas()
        if not cats_only:
            return df
        if columns is not None:
            df = df[columns]
        df.index.name = "labels"
        df.reset_index(drop=False, inplace=True)
        return df
    return table


def clean_worker_cache(name=None):
    """Utility to clean the cache dictionary for the
    current worker.  If a `name` argument is passed,
    only that element of the dictionary will be removed.
    """
    with _WORKER_CACHE_LOCK:
        try:
            worker = get_worker()
        except ValueError:
            global _WORKER_CACHE  # pylint: disable=global-statement
            if _WORKER_CACHE != {}:
                if name:
                    del _WORKER_CACHE[name]
                else:
                    del _WORKER_CACHE
                    _WORKER_CACHE = {}
            return
        if hasattr(worker, "worker_cache"):
            if name:
                del worker.worker_cache[name]
            else:
                del worker.worker_cache
        return



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

import logging

from merlin.core.dispatch import DataFrameType
from merlin.dag.base_operator import BaseOperator
from merlin.dag.selector import ColumnSelector
from merlin.schema import Schema

LOG = logging.getLogger("SelectionOp")


class SelectionOp(BaseOperator):
    """
    This operator class provides an implementation of the behavior of selection (e.g. input) nodes.
    """

    def __init__(self, selector=None):
        self.selector = selector
        super().__init__()

    def transform(self, col_selector: ColumnSelector, df: DataFrameType) -> DataFrameType:
        """Simply returns the selected output columns from the input dataframe

        The main functionality of this operator has to do with computing the schemas
        for selection nodes in the Workflow graph, so very little has to happen in the
        `transform` method.

        Parameters
        -----------
        columns: list of str or list of list of str
            The columns to apply this operator to
        df: Dataframe
            A pandas or cudf dataframe that this operator will work on

        Returns
        -------
        DataFrame
            Returns a transformed dataframe for this operator
        """
        selector = col_selector or self.selector
        return super()._get_columns(df, selector)

    def compute_input_schema(
        self,
        root_schema: Schema,
        parents_schema: Schema,
        deps_schema: Schema,
        selector: ColumnSelector,
    ) -> Schema:
        """
        Return the schemas of columns

        Parameters
        ----------
        root_schema : Schema
            Schema of the columns from the input dataset
        parents_schema : Schema
            Schema of the columns from the parent nodes
        deps_schema : Schema
            Schema of the columns from the dependency nodes
        selector : ColumnSelector
            Existing column selector for this node in the graph (often None)

        Returns
        -------
        Schema
            Schema of selected columns from input, parents, and dependencies
        """
        upstream_schema = root_schema + parents_schema + deps_schema
        return upstream_schema.apply(self.selector)

    def compute_output_schema(
        self,
        input_schema: Schema,
        col_selector: ColumnSelector,
        prev_output_schema: Schema = None,
    ) -> Schema:
        """Given a set of schemas and a column selector for the input columns,
        returns a set of schemas for the transformed columns this operator will produce

        Parameters
        -----------
        input_schema: Schema
            The schemas of the columns to apply this operator to
        col_selector: ColumnSelector
            The column selector to apply to the input schema
        Returns
        -------
        Schema
            The schemas of the columns produced by this operator
        """
        selector = col_selector or self.selector
        return super().compute_output_schema(input_schema, selector, prev_output_schema)



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from merlin.schema import ColumnSchema, Schema, Tags
from merlin.schema.io.tensorflow_metadata import TensorflowMetadata


def test_json_serialization_with_embedded_dicts():
    # there were a couple issues with exporting schemas with embedded dictionaries or lists
    # to json:
    # https://github.com/NVIDIA-Merlin/models/issues/231
    # Verify that this works like we expect

    schema = Schema(
        [
            ColumnSchema(
                "userid",
                tags=[Tags.USER_ID, Tags.CATEGORICAL],
                properties={
                    "embedding_sizes": {"cardinality": 102987, "dimension": 2.0},
                    "list_value": [{"dict_val": 1.0}],
                },
            )
        ]
    )
    json_schema = TensorflowMetadata.from_merlin_schema(schema).to_json()
    output_schema = TensorflowMetadata.from_json(json_schema).to_merlin_schema()
    assert output_schema == schema


def test_merlin_to_proto_to_json_to_merlin():
    # we used to have an issue going from merlin -> tensorflowmetadata-> proto_text ->
    # tensorflowmetadata -> json schema would throw an error like
    # TypeError: memoryview: a bytes-like object is required, not 'dict'
    # verify that this works as expected
    schema = Schema(
        [
            ColumnSchema(
                "userid",
                tags=[Tags.USER_ID, Tags.CATEGORICAL],
                properties={
                    "num_buckets": None,
                    "freq_threshold": 0.0,
                    "domain": {"min": 0, "max": 102987},
                },
            )
        ]
    )
    tfm_schema = TensorflowMetadata.from_merlin_schema(schema)
    proto_schema = tfm_schema.to_proto_text()

    tfm_schema2 = TensorflowMetadata.from_proto_text(proto_schema)
    json_schema = tfm_schema2.to_json()

    tfm_schema2 = TensorflowMetadata.from_json(json_schema)
    output_schema = tfm_schema2.to_merlin_schema()

    assert output_schema == schema



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from fsspec.core import get_fs_token_paths

from .hugectr import HugeCTRWriter
from .parquet import CPUParquetWriter, GPUParquetWriter


def writer_factory(
    output_format,
    output_path,
    out_files_per_proc,
    shuffle,
    use_guid=False,
    bytes_io=False,
    num_threads=0,
    cpu=False,
    fns=None,
    suffix=None,
):
    if output_format is None:
        return None

    writer_cls, fs = _writer_cls_factory(output_format, output_path, cpu=cpu)
    return writer_cls(
        output_path,
        num_out_files=out_files_per_proc,
        shuffle=shuffle,
        fs=fs,
        use_guid=use_guid,
        bytes_io=bytes_io,
        num_threads=num_threads,
        cpu=cpu,
        fns=fns,
        suffix=suffix,
    )


def _writer_cls_factory(output_format, output_path, cpu=None):
    if output_format == "parquet" and cpu:
        writer_cls = CPUParquetWriter
    elif output_format == "parquet":
        writer_cls = GPUParquetWriter
    elif output_format == "hugectr":
        writer_cls = HugeCTRWriter
    else:
        raise ValueError("Output format not yet supported.")

    fs = get_fs_token_paths(output_path)[0]
    return writer_cls, fs



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import enum
import warnings

import pandas as pd
from packaging.version import Version

_IGNORE_INDEX_SUPPORTED = Version(pd.__version__) >= Version("1.3.0")


class Shuffle(enum.Enum):
    PER_PARTITION = 0
    PER_WORKER = 1
    FULL = 2


#
# Helper Function definitions
#


def _check_shuffle_arg(shuffle):
    if shuffle is None:
        return shuffle

    if isinstance(shuffle, Shuffle):
        if shuffle == Shuffle.FULL:
            raise ValueError('`shuffle="full"` is not yet supported.')
    elif shuffle is True:
        shuffle = Shuffle.PER_WORKER
        warnings.warn("`shuffle=True` is deprecated. Using `PER_WORKER`.", DeprecationWarning)
    elif shuffle is False:
        shuffle = None
    else:
        raise ValueError(f"`shuffle={shuffle}` not recognized.")
    return shuffle


def shuffle_df(df, size=None, keep_index=False):
    """Shuffles a DataFrame, returning a new dataframe with randomly
    ordered rows"""
    size = size or len(df)
    if isinstance(df, pd.DataFrame):
        if _IGNORE_INDEX_SUPPORTED:
            return df.sample(n=size, ignore_index=not keep_index)
        else:
            # Pandas<1.3.0
            if keep_index:
                return df.sample(n=size)
            return df.sample(n=size).reset_index(drop=True)
    else:
        return df.sample(n=size, keep_index=keep_index)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import errno
import os
import shutil
import subprocess
import sys

from natsort import natsorted
from recommonmark.parser import CommonMarkParser

sys.path.insert(0, os.path.abspath("../../"))

repodir = os.path.abspath(os.path.join(__file__, r"../../.."))
gitdir = os.path.join(repodir, r".git")

# -- Project information -----------------------------------------------------

project = "Merlin Core"
copyright = "2022, NVIDIA"  # pylint: disable=redefined-builtin
author = "NVIDIA"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_multiversion",
    "sphinx_rtd_theme",
    "recommonmark",
    "sphinx_markdown_tables",
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
]


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

source_parsers = {".md": CommonMarkParser}
source_suffix = [".rst", ".md"]

if os.path.exists(gitdir):
    tag_refs = subprocess.check_output(["git", "tag", "-l", "v*"]).decode("utf-8").split()
    tag_refs = natsorted(tag_refs)[-6:]
    smv_tag_whitelist = r"^(" + r"|".join(tag_refs) + r")$"
else:
    smv_tag_whitelist = r"^v.*$"

smv_branch_whitelist = r"^main$"


intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "cudf": ("https://docs.rapids.ai/api/cudf/stable/", None),
    "distributed": ("https://distributed.dask.org/en/latest/", None),
}

autodoc_inherit_docstrings = False
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "show-inheritance": False,
    "member-order": "bysource",
}

autosummary_generate = True


def copy_files(src: str):
    """
    src_dir: A path, specified as relative to the
             docs/source directory in the repository.
             The source can be a directory or a file.
             Sphinx considers all directories as relative
             to the docs/source directory.

             TIP: Add these paths to the .gitignore file.
    """
    src_path = os.path.abspath(src)
    if not os.path.exists(src_path):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), src_path)
    out_path = os.path.basename(src_path)
    out_path = os.path.abspath("{}/".format(out_path))

    print(
        r"Copying source documentation from: {}".format(src_path),
        file=sys.stderr,
    )
    print(r"  ...to destination: {}".format(out_path), file=sys.stderr)

    if os.path.exists(out_path) and os.path.isdir(out_path):
        shutil.rmtree(out_path, ignore_errors=True)
    if os.path.exists(out_path) and os.path.isfile(out_path):
        os.unlink(out_path)

    if os.path.isdir(src_path):
        shutil.copytree(src_path, out_path)
    else:
        shutil.copyfile(src_path, out_path)


copy_files(r"../../README.md")



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from merlin.core.utils import Distributed, Serial, global_dask_client, set_dask_client

try:
    import cudf

    _CPU = [True, False]
except ImportError:
    _CPU = [True]
    cudf = None
_HAS_GPU = cudf is not None


@pytest.mark.parametrize("cpu", _CPU)
def test_serial_context(client, cpu):

    # Set distributed client
    set_dask_client(client=client)
    assert global_dask_client() == client

    # Check that the global dask client
    # becomes None in a `with Serial()` block
    with Serial():
        assert global_dask_client() is None

    # Global client should revert outside
    # the `with Serial()` block
    assert global_dask_client() == client


@pytest.mark.parametrize("cpu", [True, False])
@pytest.mark.parametrize("nested_serial", _CPU)
def test_nvt_distributed(cpu, nested_serial):

    if cpu:
        distributed = pytest.importorskip("distributed")
        cluster_type = "cpu"
        cluster_cls = distributed.LocalCluster
    else:
        dask_cuda = pytest.importorskip("dask_cuda")
        cluster_type = "cuda"
        cluster_cls = dask_cuda.LocalCUDACluster

    # Set the global client to None
    set_dask_client(client=None)
    assert global_dask_client() is None

    # Check that a new local cluster is deployed within
    # a `with Distributed()` block
    with Distributed(cluster_type=cluster_type, n_workers=1, force_new=True) as dist:
        assert dist.client is not None
        assert global_dask_client() == dist.client
        assert len(dist.cluster.workers) == 1
        assert isinstance(dist.cluster, cluster_cls)

        # Check that we can nest a `with Serial()` block
        # inside a `with Distributed()` block
        if nested_serial:
            with Serial():
                assert global_dask_client() is None
            assert global_dask_client() == dist.client

    # Global client should revert to None outside
    # the `with Distributed()` block
    assert global_dask_client() is None


@pytest.mark.parametrize("cpu", _CPU)
def test_nvt_distributed_force(client, cpu):

    if cpu:
        distributed = pytest.importorskip("distributed")
        cluster_type = "cpu"
        cluster_cls = distributed.LocalCluster
    else:
        dask_cuda = pytest.importorskip("dask_cuda")
        cluster_type = "cuda"
        cluster_cls = dask_cuda.LocalCUDACluster

    # Set distributed client
    set_dask_client(client=client)
    assert global_dask_client() == client

    # Check that a new local cluster is deployed within
    # a `with Distributed()` block. Since we are using
    # `force_new=True`, the new cluster should NOT be
    # the same as the original `client`.
    with Distributed(cluster_type=cluster_type, force_new=True, n_workers=1) as dist:
        assert dist.client != client
        assert global_dask_client() == dist.client
        assert len(dist.cluster.workers) == 1
        assert isinstance(dist.cluster, cluster_cls)

    # We should revert to the original client
    # outside the `with Distributed()` block
    assert global_dask_client() == client

    # Check that the default behavior is to avoid
    # deploying a new cluster (and warning the user)
    # if an existing client is detected
    with pytest.warns(UserWarning):
        with Distributed(cluster_type=cluster_type, n_workers=1) as dist:
            assert dist.client == client
            assert global_dask_client() == dist.client

    # We should revert to the original client
    # outside the `with Distributed()` block
    assert global_dask_client() == client



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from merlin.core.dispatch import DataFrameType
from merlin.dag.base_operator import BaseOperator
from merlin.dag.selector import ColumnSelector


class SubsetColumns(BaseOperator):
    """
    This operator class provides an implementation for the `[]` operator
    used in constructing graphs.
    """

    def __init__(self, label=None):
        self._label = label or self.__class__.__name__
        super().__init__()

    def transform(self, col_selector: ColumnSelector, df: DataFrameType) -> DataFrameType:
        """Simply returns the selected output columns from the input dataframe

        The main functionality of this operator has to do with computing the schemas
        for `-` nodes in the Workflow graph, so very little has to happen in the
        `transform` method.

        Parameters
        -----------
        columns: list of str or list of list of str
            The columns to apply this operator to
        df: Dataframe
            A pandas or cudf dataframe that this operator will work on

        Returns
        -------
        DataFrame
            Returns a transformed dataframe for this operator
        """
        return super()._get_columns(df, col_selector)

    @property
    def label(self) -> str:
        return self._label



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import collections

import dask
import pandas as pd
from dask.base import tokenize
from dask.dataframe.core import _concat, new_dd_object
from dask.delayed import Delayed
from dask.highlevelgraph import HighLevelGraph

from merlin.core.dispatch import annotate
from merlin.core.utils import ensure_optimize_dataframe_graph, global_dask_client
from merlin.io.worker import clean_worker_cache, get_worker_cache

from .shuffle import Shuffle
from .writer_factory import _writer_cls_factory, writer_factory


class DaskSubgraph:
    """Simple container for a Dask subgraph

    Parameters
    ----------
    full_graph: dask.HighLevelGraph
        Full graph for some DataFrame collection.
    keys: set
        Subset of collection keys required for this subgraph.
        This set will be used to cull all unrelated tasks from
        the full graph during initialization.
    """

    def __init__(self, full_graph, name, parts):
        self.name = name
        self.parts = parts
        self.subgraph = full_graph.cull({(self.name, part) for part in self.parts})

    def __getitem__(self, part):
        key = (self.name, part)
        dsk = self.subgraph.cull({key})
        return dask.get(dsk, key)


@annotate("write_output_partition", color="green", domain="merlin_python")
def _write_output_partition(
    df,
    processed_path,
    shuffle,
    out_files_per_proc,
    fs,
    cat_names,
    cont_names,
    label_names,
    output_format,
    num_threads,
    cpu,
    suffix,
):
    df_size = len(df)
    out_files_per_proc = out_files_per_proc or 1

    # Get cached writer (or create/cache a new one)
    with get_worker_cache("writer") as writer_cache:
        writer = writer_cache.get(processed_path, None)
        if writer is None:
            writer = writer_factory(
                output_format,
                processed_path,
                out_files_per_proc,
                shuffle,
                use_guid=True,
                bytes_io=(shuffle == Shuffle.PER_WORKER),
                num_threads=num_threads,
                cpu=cpu,
                suffix=suffix,
            )
            writer.set_col_names(labels=label_names, cats=cat_names, conts=cont_names)
            writer_cache[processed_path] = writer

        # Add data
        writer.add_data(df)

    return df_size


def _get_partition_groups(df, partition_cols, fs, output_path, filename):
    # Logic copied from cudf/cudf/io/parquet.py
    df = df.sort_values(partition_cols)
    divisions = df[partition_cols].drop_duplicates(ignore_index=True)
    splits = df[partition_cols].searchsorted(divisions, side="left")
    splits = splits.tolist() + [len(df[partition_cols])]

    subgroups, fns = [], []
    for i in range(0, len(splits) - 1):
        sub_df = df.iloc[splits[i] : splits[i + 1]].copy(deep=False)
        if sub_df is None or len(sub_df) == 0:
            continue
        keys = tuple(sub_df[col].iloc[0] for col in partition_cols)
        if not isinstance(keys, tuple):
            keys = (keys,)
        subdir = fs.sep.join(
            [
                "{colname}={value}".format(colname=name, value=val)
                for name, val in zip(partition_cols, keys)
            ]
        )
        prefix = fs.sep.join([output_path, subdir])
        fs.mkdirs(prefix, exist_ok=True)
        fns.append(fs.sep.join([subdir, filename]))

        write_df = sub_df.copy(deep=False)
        write_df.drop(columns=partition_cols, inplace=True)
        subgroups.append(write_df)

    return fns, subgroups


@annotate("write_partitioned", color="green", domain="merlin_python")
def _write_partitioned(
    df,
    filename,
    output_path,
    partition_cols,
    shuffle,
    fs,
    cat_names,
    cont_names,
    label_names,
    output_format,
    num_threads,
    cpu,
):
    # Logic copied from cudf/cudf/io/parquet.py
    data_cols = df.columns.drop(partition_cols)
    if len(data_cols) == 0:
        raise ValueError("No data left to save outside partition columns")

    # Partition the input data
    fns, dfs = _get_partition_groups(df, partition_cols, fs, output_path, filename)
    writer = writer_factory(
        output_format,
        output_path,
        1,
        shuffle,
        num_threads=num_threads,
        cpu=cpu,
        fns=fns,
    )
    writer.set_col_names(labels=label_names, cats=cat_names, conts=cont_names)

    # Add the partitioned data
    writer.add_data(dfs)

    # Return metadata and row-count in dict
    return writer.close()


@annotate("write_subgraph", color="green", domain="merlin_python")
def _write_subgraph(
    subgraph,
    fns,
    output_path,
    shuffle,
    fs,
    cat_names,
    cont_names,
    label_names,
    output_format,
    num_threads,
    cpu,
    suffix,
):

    fns = fns if isinstance(fns, (tuple, list)) else (fns,)
    writer = writer_factory(
        output_format,
        output_path,
        len(fns),
        shuffle,
        bytes_io=(shuffle == Shuffle.PER_WORKER),
        num_threads=num_threads,
        cpu=cpu,
        fns=[fn + suffix for fn in fns],
    )
    writer.set_col_names(labels=label_names, cats=cat_names, conts=cont_names)

    # Add data
    num_rows = 0
    for part in subgraph.parts:
        table = subgraph[part]
        writer.add_data(table)
        num_rows += len(table)
        del table

    # Return metadata and row-count in dict
    return writer.close()


def _write_metadata_files(md_list, output_path, output_format, cpu, schema):

    # Separate and merge metadata
    general_md = []
    special_md = []
    for md in md_list:
        general_md.append(md[0])
        special_md.append(md[1])
    general_md = _merge_general_metadata(general_md)
    special_md = dict(collections.ChainMap(*special_md))

    # Write metadata files
    if not isinstance(output_path, str):
        output_path = str(output_path)
    wc, fs = _writer_cls_factory(output_format, output_path, cpu)
    wc.write_general_metadata(general_md, fs, output_path, schema)
    wc.write_special_metadata(special_md, fs, output_path)


def _simple_shuffle(ddf, plan):

    # Construct graph for a simple shuffle
    token = tokenize(ddf, plan)
    name = "shuffled-" + token
    final_tasks = collections.defaultdict(list)
    ignore_index = True
    for i, p in enumerate(plan):
        final_tasks[(name, p)].append((ddf._name, i))
    dsk = {k: (_concat, v, ignore_index) for k, v in final_tasks.items()}

    # Convert to a DataFrame collection
    graph = HighLevelGraph.from_collections(name, dsk, dependencies=[ddf])
    divisions = [None] * (len(dsk) + 1)
    return new_dd_object(graph, name, ddf._meta, divisions)


def _ddf_to_dataset(
    ddf,
    fs,
    output_path,
    shuffle,
    file_partition_map,
    out_files_per_proc,
    cat_names,
    cont_names,
    label_names,
    output_format,
    num_threads,
    cpu,
    suffix="",
    partition_on=None,
    schema=None,
):

    # Construct graph for Dask-based dataset write
    token = tokenize(
        ddf, shuffle, out_files_per_proc, cat_names, cont_names, label_names, suffix, partition_on
    )
    name = "write-processed-" + token
    write_name = name + "-partition" + token

    # Check that the data is in the correct place
    assert isinstance(ddf._meta, pd.DataFrame) is cpu

    dsk = {}
    task_list = []
    if partition_on:
        # Use hive partitioning to write the data
        cached_writers = False
        for idx in range(ddf.npartitions):
            task_list.append((write_name, idx))
            dsk[task_list[-1]] = (
                _write_partitioned,
                (ddf._name, idx),
                f"part.{idx}{suffix}",
                output_path,
                partition_on,
                shuffle,
                fs,
                cat_names,
                cont_names,
                label_names,
                output_format,
                num_threads,
                cpu,
            )
        dsk[name] = (
            _write_metadata_files,
            task_list,
            output_path,
            output_format,
            cpu,
            schema,
        )
    elif file_partition_map is not None:
        # Use specified mapping of data to output files
        cached_writers = False
        full_graph = ddf.dask
        for fns, parts in file_partition_map.items():
            # Isolate subgraph for this output file
            subgraph = DaskSubgraph(full_graph, ddf._name, parts)
            task_list.append((write_name, str(fns)))
            dsk[task_list[-1]] = (
                _write_subgraph,
                subgraph,
                fns,
                output_path,
                shuffle,
                fs,
                cat_names,
                cont_names,
                label_names,
                output_format,
                num_threads,
                cpu,
                suffix,
            )
        dsk[name] = (
            _write_metadata_files,
            task_list,
            output_path,
            output_format,
            cpu,
            schema,
        )
    else:
        cached_writers = True
        for idx in range(ddf.npartitions):
            key = (write_name, idx)
            dsk[key] = (
                _write_output_partition,
                (ddf._name, idx),
                output_path,
                shuffle,
                out_files_per_proc,
                fs,
                cat_names,
                cont_names,
                label_names,
                output_format,
                num_threads,
                cpu,
                suffix,
            )
            task_list.append(key)
        dsk[name] = (lambda x: x, task_list)

    graph = ensure_optimize_dataframe_graph(
        dsk=HighLevelGraph.from_collections(name, dsk, dependencies=[ddf]),
        keys=[name],
    )
    out = Delayed(name, graph)

    # Trigger write execution
    client = global_dask_client()
    if client:
        out = client.compute(out).result()
    else:
        out = dask.compute(out, scheduler="synchronous")[0]

    if cached_writers:
        # Follow-up Shuffling and _metadata creation
        _finish_dataset(client, ddf, output_path, fs, output_format, cpu, schema)


def _finish_dataset(client, ddf, output_path, fs, output_format, cpu, schema):
    # Finish data writing
    if client:
        client.cancel(ddf)
        ddf = None
        out = client.run(_worker_finish, output_path)

        general_md = []
        special_md = []
        for (gen, spec) in out.values():
            general_md.append(gen)
            if spec:
                special_md.append(spec)

        general_md = _merge_general_metadata(general_md)
        special_md = dict(collections.ChainMap(*special_md))
    else:
        ddf = None
        general_md, special_md = _worker_finish(output_path)

    # Write metadata on client
    if not isinstance(output_path, str):
        output_path = str(output_path)

    wc, fs = _writer_cls_factory(output_format, output_path, cpu)
    wc.write_general_metadata(general_md, fs, output_path, schema)
    wc.write_special_metadata(special_md, fs, output_path)

    # Clean writer caches
    if client:
        client.run(clean_worker_cache, "writer")
    else:
        clean_worker_cache("writer")


def _worker_finish(processed_path):
    general_md, special_md = {}, {}
    with get_worker_cache("writer") as writer_cache:
        writer = writer_cache.get(processed_path, None)
        if writer:
            general_md, special_md = writer.close()

    return general_md, special_md


def _merge_general_metadata(meta_list):
    """Combine list of "general" metadata dicts into
    a single dict
    """
    if not meta_list:
        return {}
    meta = None
    for md in meta_list:
        if meta:
            if "data_paths" in md:
                meta["data_paths"] += md["data_paths"]
            if "file_stats" in md:
                meta["file_stats"] += md["file_stats"]
        else:
            meta = md.copy()
            if "data_paths" not in meta:
                meta["data_paths"] = []
            if "file_stats" not in meta:
                meta["file_stats"] = []
    return meta



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
from collections import deque

from merlin.dag.node import (
    Node,
    _combine_schemas,
    iter_nodes,
    postorder_iter_nodes,
    preorder_iter_nodes,
)
from merlin.schema import Schema

LOG = logging.getLogger("merlin")


class Graph:
    def __init__(self, output_node: Node):
        self.output_node = output_node

    @property
    def input_dtypes(self):
        if self.input_schema:
            return {
                name: col_schema.dtype
                for name, col_schema in self.input_schema.column_schemas.items()
            }
        else:
            return {}

    @property
    def output_dtypes(self):
        if self.output_schema:
            return {
                name: col_schema.dtype
                for name, col_schema in self.output_schema.column_schemas.items()
            }
        else:
            return {}

    @property
    def column_mapping(self):
        nodes = preorder_iter_nodes(self.output_node)
        column_mapping = self.output_node.column_mapping
        for node in list(nodes)[1:]:
            node_map = node.column_mapping
            for output_col, input_cols in column_mapping.items():
                early_inputs = []
                for input_col in input_cols:
                    early_inputs += node_map.get(input_col, [input_col])
                column_mapping[output_col] = early_inputs

        return column_mapping

    def construct_schema(self, root_schema: Schema, preserve_dtypes=False) -> "Graph":
        nodes = list(postorder_iter_nodes(self.output_node))

        self._compute_node_schemas(root_schema, nodes, preserve_dtypes)
        self._validate_node_schemas(root_schema, nodes, preserve_dtypes)

        return self

    def _compute_node_schemas(self, root_schema, nodes, preserve_dtypes=False):
        for node in nodes:
            node.compute_schemas(root_schema, preserve_dtypes=preserve_dtypes)

    def _validate_node_schemas(self, root_schema, nodes, strict_dtypes=False):
        for node in nodes:
            node.validate_schemas(root_schema, strict_dtypes=strict_dtypes)

    @property
    def input_schema(self):
        # leaf_node input and output schemas are the same (aka selection)
        return _combine_schemas(self.leaf_nodes)

    @property
    def leaf_nodes(self):
        return [node for node in postorder_iter_nodes(self.output_node) if not node.parents]

    @property
    def output_schema(self):
        return self.output_node.output_schema

    def _input_columns(self):
        input_cols = []
        for node in iter_nodes([self.output_node]):
            upstream_output_cols = []

            for upstream_node in node.parents_with_dependencies:
                upstream_output_cols += upstream_node.output_columns.names

            upstream_output_cols = _get_unique(upstream_output_cols)
            input_cols += list(set(node.input_columns.names) - set(upstream_output_cols))

        return _get_unique(input_cols)

    def remove_inputs(self, to_remove):
        """
        Removes columns from a Graph

        Starting at the leaf nodes, trickle down looking for columns to remove,
        when found remove but then must propagate the removal of any other
        output columns derived from that column.

        Parameters
        -----------
        graph : Graph
            The graph to remove columns from
        to_remove : array_like
            A list of input column names to remove from the graph

        Returns
        -------
        Graph
            The same graph with columns removed
        """
        nodes_to_process = deque([(node, to_remove) for node in self.leaf_nodes])

        while nodes_to_process:
            node, columns_to_remove = nodes_to_process.popleft()

            if node.input_schema and len(node.input_schema):
                output_columns_to_remove = node.remove_inputs(columns_to_remove)

                for child in node.children:
                    nodes_to_process.append((child, to_remove + output_columns_to_remove))

                    if not len(node.input_schema):
                        node.remove_child(child)

            if not node.input_schema or not len(node.input_schema):
                for parent in node.parents:
                    parent.remove_child(node)
                for dependency in node.dependencies:
                    dependency.remove_child(node)
                del node

        return self

    @classmethod
    def get_nodes_by_op_type(cls, nodes, op_type):
        return set(node for node in iter_nodes(nodes) if isinstance(node.op, op_type))


def _get_schemaless_nodes(nodes):
    schemaless_nodes = []
    for node in iter_nodes(nodes):
        if node.input_schema is None:
            schemaless_nodes.append(node)

    return set(schemaless_nodes)


def _get_unique(cols):
    # Need to preserve order in unique-column list
    return list({x: x for x in cols}.keys())



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import io
from threading import Thread

# Check if fsspec.parquet module is available
import fsspec
import numpy as np
from packaging.version import Version
from pyarrow import parquet as pq

if Version(fsspec.__version__) > Version("2021.11.0"):
    import fsspec.parquet as fsspec_parquet
else:
    fsspec_parquet = None

try:
    import cudf
    from cudf.core.column import as_column, build_categorical_column

    if fsspec_parquet and (Version(cudf.__version__) < Version("21.12.0")):
        # This version of cudf does not support
        # reads from python-file objects. Avoid
        # using the `fsspec.parquet` module
        fsspec_parquet = None
except ImportError:
    cudf = None

#
# Parquet-Specific Utilities
#


def _optimized_read_partition_remote(
    fs, pieces, columns, index, categories=(), partitions=(), **kwargs
):
    # This is a specialized version of `CudfEngine.read_partition`
    # for remote filesystems. This implementation is intended to
    # replace the upstream `read_partition` classmethod until
    # remote-filesystem handling is optimized in cudf/dask-cudf

    if columns is not None:
        columns = list(columns)
    if isinstance(index, list):
        columns += index

    # Check that this is a single-piece read on a non-local filesystem
    if not isinstance(pieces, list):
        pieces = [pieces]
    if len(pieces) > 1:
        raise ValueError(
            "The `_custom_read_partition` code path is not designed to "
            "handle a multi-element `pieces` argument."
        )
    if cudf.utils.ioutils._is_local_filesystem(fs):
        raise ValueError(
            "The `_custom_read_partition` code path is not intended "
            "for use on local filesystems."
        )

    # Unpack contents of the single piece
    if isinstance(pieces[0], str):
        path = pieces[0]
        row_group = None
        partition_keys = []
    else:
        (path, row_group, partition_keys) = pieces[0]

    # Call optimized read utility
    df = _optimized_read_remote(path, row_group, columns, fs, **kwargs)

    #
    # Code below is directly copied from cudf-21.08
    #

    if index and (index[0] in df.columns):
        df = df.set_index(index[0])
    elif index is False and set(df.index.names).issubset(columns):
        # If index=False, we need to make sure all of the
        # names in `columns` are actually in `df.columns`
        df.reset_index(inplace=True)

    if partition_keys:
        if partitions is None:
            raise ValueError("Must pass partition sets")
        for i, (name, index2) in enumerate(partition_keys):
            categories = [val.as_py() for val in partitions.levels[i].dictionary]

            col = as_column(index2).as_frame().repeat(len(df))._data[None]
            df[name] = build_categorical_column(
                categories=categories,
                codes=as_column(col.base_data, dtype=col.dtype),
                size=col.size,
                offset=col.offset,
                ordered=False,
            )

    return df


def _optimized_read_remote(path, row_groups, columns, fs, **kwargs):

    if row_groups is not None and not isinstance(row_groups, list):
        row_groups = [row_groups]

    # Handle various key-word argument options
    user_kwargs = kwargs.copy()
    read_kwargs = user_kwargs.pop("read", {})
    strings_to_cats = read_kwargs.get(
        "strings_to_categorical",
        user_kwargs.pop("strings_to_categorical", False),
    )

    if fsspec_parquet:
        # For newer versions of fsspec/cudf, we should use
        # the `fsspec.parquet.open_parquet_file` function.
        # This approach will result in fast data transfer,
        # and will typically reduce host-memory usage.
        with fsspec_parquet.open_parquet_file(
            path,
            fs=fs,
            columns=columns,
            row_groups=row_groups,
            engine="pyarrow",
        ) as f:
            return cudf.read_parquet(
                f,
                engine="cudf",
                columns=columns,
                row_groups=row_groups,
                strings_to_categorical=strings_to_cats,
                use_python_file_object=True,
                **read_kwargs,
            )
    else:
        # Get byte-ranges that are known to contain the
        # required data for this read
        byte_ranges, footer, file_size = _get_parquet_byte_ranges(
            path, row_groups, columns, fs, **user_kwargs
        )

        return cudf.read_parquet(
            # Transfer the required bytes with fsspec
            io.BytesIO(
                _fsspec_data_transfer(
                    path,
                    fs,
                    byte_ranges=byte_ranges,
                    footer=footer,
                    file_size=file_size,
                    add_par1_magic=True,
                    **user_kwargs,
                )
            ),
            engine="cudf",
            columns=columns,
            row_groups=row_groups,
            strings_to_categorical=strings_to_cats,
            **read_kwargs,
        )


def _get_parquet_byte_ranges(
    path,
    rgs,
    columns,
    fs,
    bytes_per_thread=256_000_000,
    **kwargs,
):
    # The purpose of this utility is to return a list
    # of byte ranges (in path) that are known to contain
    # the data needed to read `columns` and `rgs`

    # Step 0 - Get size of file
    file_size = fs.size(path)

    # Return early if the file is too small to merit
    # optimized data transfer
    if file_size <= bytes_per_thread:
        return None, None, file_size

    # Step 1 - Get 32 KB from tail of file.
    #
    # This "sample size" can be tunable, but should
    # always be >= 8 bytes (so we can read the footer size)
    tail_size = 32_000
    footer_sample = fs.tail(path, tail_size)

    # Step 2 - Read the footer size and re-read a larger
    #          tail if necessary
    footer_size = int.from_bytes(footer_sample[-8:-4], "little")
    if tail_size < (footer_size + 8):
        footer_sample = fs.tail(path, footer_size + 8)

    # Step 3 - Collect required byte ranges
    byte_ranges = []
    md = pq.ParquetFile(io.BytesIO(footer_sample)).metadata
    for r in range(md.num_row_groups):
        # Skip this row-group if we are targeting
        # specific row-groups
        if rgs is None or r in rgs:
            row_group = md.row_group(r)
            for c in range(row_group.num_columns):
                column = row_group.column(c)
                name = column.path_in_schema
                # Skip this column if we are targeting
                # specific columns, and this name is not
                # in the list.
                #
                # Note that `column.path_in_schema` may
                # modify the column name for list and struct
                # columns. For example, a column named "a"
                # may become "a.list.element"
                split_name = name.split(".")[0]
                if columns is None or name in columns or split_name in columns:
                    file_offset0 = column.dictionary_page_offset
                    if file_offset0 is None:
                        file_offset0 = column.data_page_offset
                    num_bytes = column.total_compressed_size
                    byte_ranges.append((file_offset0, num_bytes))

    return byte_ranges, footer_sample, file_size


#
# General Fsspec Data-transfer Optimization Code
#


def _fsspec_data_transfer(
    path_or_fob,
    fs,
    byte_ranges=None,
    footer=None,
    file_size=None,
    add_par1_magic=None,
    bytes_per_thread=256_000_000,
    max_gap=64_000,
    mode="rb",
    **kwargs,
):

    # Calculate total file size
    file_size = file_size or fs.size(path_or_fob)

    # Check if a direct read makes the most sense
    if not byte_ranges and bytes_per_thread >= file_size:
        return fs.open(path_or_fob, mode=mode, cache_type="none").read()

    # Threaded read into "dummy" buffer
    buf = np.zeros(file_size, dtype="b")
    if byte_ranges:

        # Optimize/merge the ranges
        byte_ranges = _merge_ranges(
            byte_ranges,
            max_block=bytes_per_thread,
            max_gap=max_gap,
        )

        # Call multi-threaded data transfer of
        # remote byte-ranges to local buffer
        _read_byte_ranges(
            path_or_fob,
            byte_ranges,
            buf,
            fs,
            **kwargs,
        )

        # Add Header & Footer bytes
        if footer is not None:
            footer_size = len(footer)
            buf[-footer_size:] = np.frombuffer(footer[-footer_size:], dtype="b")

        # Add parquet magic bytes (optional)
        if add_par1_magic:
            buf[:4] = np.frombuffer(b"PAR1", dtype="b")
            if footer is None:
                buf[-4:] = np.frombuffer(b"PAR1", dtype="b")

    else:
        byte_ranges = [
            (b, min(bytes_per_thread, file_size - b)) for b in range(0, file_size, bytes_per_thread)
        ]
        _read_byte_ranges(
            path_or_fob,
            byte_ranges,
            buf,
            fs,
            **kwargs,
        )

    # Wrap in BytesIO since cudf will sometimes use
    # pyarrow to parse the metadata (and pyarrow
    # cannot read from a bytes object)
    return io.BytesIO(buf)


def _merge_ranges(byte_ranges, max_block=256_000_000, max_gap=64_000):
    # Simple utility to merge small/adjacent byte ranges
    new_ranges = []
    if not byte_ranges:
        # Early return
        return new_ranges

    offset, size = byte_ranges[0]
    for (new_offset, new_size) in byte_ranges[1:]:
        gap = new_offset - (offset + size)
        if gap > max_gap or (size + new_size + gap) > max_block:
            # Gap is too large or total read is too large
            new_ranges.append((offset, size))
            offset = new_offset
            size = new_size
            continue
        size += new_size + gap
    new_ranges.append((offset, size))
    return new_ranges


def _assign_block(fs, path_or_fob, local_buffer, offset, nbytes):
    with fs.open(path_or_fob, mode="rb", cache_type="none") as fob:
        fob.seek(offset)
        local_buffer[offset : offset + nbytes] = np.frombuffer(
            fob.read(nbytes),
            dtype="b",
        )


def _read_byte_ranges(
    path_or_fob,
    ranges,
    local_buffer,
    fs,
    **kwargs,
):

    workers = []
    for (offset, nbytes) in ranges:
        if len(ranges) > 1:
            workers.append(
                Thread(target=_assign_block, args=(fs, path_or_fob, local_buffer, offset, nbytes))
            )
            workers[-1].start()
        else:
            _assign_block(fs, path_or_fob, local_buffer, offset, nbytes)

    for worker in workers:
        worker.join()



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from merlin.dag import BaseOperator
from merlin.dag.node import Node
from merlin.dag.ops.selection import SelectionOp
from merlin.dag.selector import ColumnSelector
from merlin.schema import ColumnSchema, Schema, Tags


def test_constructor_works_with_single_strings_and_lists():
    selector1 = ColumnSelector("a")
    assert selector1._names == ["a"]

    selector2 = ColumnSelector(["a", "b", "c"])
    assert selector2._names == ["a", "b", "c"]


def test_constructor_no_names():
    selector2 = ColumnSelector(["a", "b", "c"])
    selector3 = ColumnSelector(subgroups=selector2)
    assert selector3.subgroups[0] == selector2


def test_constructor_works_with_single_subgroups_and_lists():
    selector1 = ColumnSelector([], subgroups=ColumnSelector("a"))
    assert isinstance(selector1.subgroups, list)
    assert selector1.subgroups[0] == ColumnSelector("a")

    selector2 = ColumnSelector([], subgroups=ColumnSelector(["a", "b", "c"]))
    assert isinstance(selector2.subgroups, list)
    assert selector2.subgroups[0] == ColumnSelector(["a", "b", "c"])


def test_constructor_too_many_level():
    with pytest.raises(AttributeError) as exc_info:
        ColumnSelector(
            ["h", "i"],
            subgroups=ColumnSelector(names=["b"], subgroups=ColumnSelector(["a"])),
        )
    assert "Too many" in str(exc_info.value)


def test_constructor_rejects_workflow_nodes():
    group = Node(ColumnSelector(["a"]))

    with pytest.raises(TypeError) as exception_info:
        ColumnSelector(group)

    assert "Node" in str(exception_info.value)

    with pytest.raises(ValueError) as exception_info:
        ColumnSelector(["a", "b", group])

    assert "Node" in str(exception_info.value)


def test_constructor_creates_subgroups_from_nesting():
    selector = ColumnSelector(["a", "b", "c", ["d", "e", "f"]])

    assert selector._names == ["a", "b", "c"]
    assert selector.subgroups == [ColumnSelector(["d", "e", "f"])]

    selector = ColumnSelector(["a", "b", "c", ("d", "e", "f")])

    assert selector._names == ["a", "b", "c"]
    assert selector.subgroups == [ColumnSelector(["d", "e", "f"])]


def test_lists_containing_selectors_create_subgroups():
    selector = ColumnSelector(["a", "b", "c", ColumnSelector(["d", "e", "f"])])

    assert len(selector.subgroups) == 1
    assert selector.subgroups == [ColumnSelector(["d", "e", "f"])]
    assert selector.grouped_names == ["a", "b", "c", ("d", "e", "f")]


def test_names_returns_flat_list():
    selector = ColumnSelector(["a", "b", "c"], [ColumnSelector(["d", "e", "f"])])

    assert selector.names == ["a", "b", "c", "d", "e", "f"]


def test_grouped_names_returns_nested_list():
    selector = ColumnSelector(["a", "b", "c"], [ColumnSelector(["d", "e", "f"])])

    assert selector.grouped_names == ["a", "b", "c", ("d", "e", "f")]


def test_returned_names_are_unique():
    selector = ColumnSelector(["a", "b", "a"])
    assert selector.names == ["a", "b"]

    selector = ColumnSelector([("a", "b"), ("a", "b")])
    assert selector.grouped_names == [("a", "b")]


def test_addition_combines_names_and_subgroups():
    selector1 = ColumnSelector(["a", "b", "c", ["d", "e", "f"]])
    selector2 = ColumnSelector(["g", "h", "i", ["j", "k", "l"]])
    combined = selector1 + selector2

    assert combined._names == ["a", "b", "c", "g", "h", "i"]
    assert combined.subgroups[0]._names == ["d", "e", "f"]
    assert combined.subgroups[1]._names == ["j", "k", "l"]
    assert len(combined.subgroups) == 2


def test_addition_works_with_strings():
    selector = ColumnSelector(["a", "b", "c", "d", "e", "f"])
    combined = selector + "g"

    assert combined.names == ["a", "b", "c", "d", "e", "f", "g"]


def test_addition_works_with_lists_of_strings():
    selector = ColumnSelector(["a", "b", "c"])
    combined = selector + ["d", "e", "f"]

    assert combined.names == ["a", "b", "c", "d", "e", "f"]


def test_addition_works_with_none():
    selector = ColumnSelector(["a", "b", "c"])
    combined = selector + None

    assert combined.names == ["a", "b", "c"]


def test_rshift_operator_onto_selector_creates_selection_node():
    selector = ColumnSelector(["a", "b", "c"])
    operator = BaseOperator()

    output_node = selector >> operator

    assert isinstance(output_node, Node)
    assert isinstance(output_node.op, BaseOperator)
    assert output_node._selector is None
    assert len(output_node.parents) == 1

    assert isinstance(output_node.parents[0], Node)
    assert isinstance(output_node.parents[0].op, SelectionOp)
    assert output_node.parents[0]._selector == selector
    assert len(output_node.parents[0].parents) == 0


def test_construct_column_selector_with_tags():
    target_tags = [Tags.CATEGORICAL, "custom"]
    selector = ColumnSelector(tags=target_tags)
    assert selector.tags == target_tags


def test_returned_tags_are_unique():
    selector = ColumnSelector(tags=["a", "b", "a"])
    assert selector.tags == ["a", "b"]


def test_addition_combines_tags():
    selector1 = ColumnSelector(tags=["a", "b", "c"])
    selector2 = ColumnSelector(tags=["g", "h", "i"])
    combined = selector1 + selector2

    assert combined.tags == ["a", "b", "c", "g", "h", "i"]


def test_addition_combines_names_and_tags():
    selector1 = ColumnSelector(["a", "b", "c"])
    selector2 = ColumnSelector(tags=["g", "h", "i"])
    combined = selector1 + selector2

    assert combined.names == ["a", "b", "c"]
    assert combined.tags == ["g", "h", "i"]


def test_addition_enum_tags():
    selector1 = ColumnSelector(tags=["a", "b", "c"])
    combined = selector1 + Tags.CATEGORICAL

    assert combined.tags == ["a", "b", "c", Tags.CATEGORICAL]

    selector2 = ColumnSelector(["a", "b", "c", ["d", "e", "f"]])
    combined = selector2 + Tags.CATEGORICAL

    assert combined._names == ["a", "b", "c"]
    assert combined.subgroups == [ColumnSelector(["d", "e", "f"])]
    assert combined.tags == [Tags.CATEGORICAL]


def test_filter():
    selector1 = ColumnSelector(["a", "b", "c"])
    sub_selector = ColumnSelector(["c"])

    result_selector = selector1.filter_columns(sub_selector)
    assert result_selector.names == ["a", "b"]

    selector2 = ColumnSelector(["a", "b", "c"], subgroups=[ColumnSelector(["a", "b", "c"])])
    result_selector = selector2.filter_columns(sub_selector)
    assert result_selector.names == ["a", "b"]
    assert result_selector.subgroups == []


def test_filter_group():
    selector = ColumnSelector([("a", "b"), ("c", "d")])

    result_selector = selector.filter_columns(ColumnSelector(["c"]))

    assert result_selector.names == ["a", "b"]
    assert result_selector.subgroups == [ColumnSelector(["a", "b"])]


def test_applying_selector_to_schema_selects_by_name():
    schema = Schema(["a", "b", "c", "d", "e"])
    selector = ColumnSelector(["a", "b"])
    result = schema.apply(selector)

    assert result == Schema(["a", "b"])

    selector = None
    result = schema.apply(selector)

    assert result == schema


def test_applying_selector_to_schema_selects_by_tags():
    schema1 = ColumnSchema("col1", tags=["a", "b", "c"])
    schema2 = ColumnSchema("col2", tags=["b", "c", "d"])

    schema = Schema([schema1, schema2])
    selector = ColumnSelector(tags=["a", "b"])
    result = schema.apply(selector)

    assert result.column_names == schema.column_names


def test_applying_selector_to_schema_selects_by_name_or_tags():
    schema1 = ColumnSchema("col1")
    schema2 = ColumnSchema("col2", tags=["b", "c", "d"])

    schema = Schema([schema1, schema2])
    selector = ColumnSelector(["col1"], tags=["a", "b"])
    result = schema.apply(selector)

    assert result.column_names == schema.column_names


def test_applying_inverse_selector_to_schema_selects_relevant_columns():
    schema = Schema(["a", "b", "c", "d", "e"])
    selector = ColumnSelector(["a", "b"])
    result = schema.apply_inverse(selector)

    assert result == Schema(["c", "d", "e"])

    selector = None
    result = schema.apply_inverse(selector)

    assert result == schema



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from merlin.core.dispatch import DataFrameType
from merlin.dag.base_operator import BaseOperator
from merlin.dag.selector import ColumnSelector
from merlin.schema import Schema


class ConcatColumns(BaseOperator):
    """
    This operator class provides an implementation for the `+` operator used in constructing graphs.
    """

    def __init__(self, label=None):
        self._label = label or self.__class__.__name__
        super().__init__()

    def compute_selector(
        self,
        input_schema: Schema,
        selector: ColumnSelector,
        parents_selector: ColumnSelector,
        dependencies_selector: ColumnSelector,
    ) -> ColumnSelector:
        """
        Combine selectors from the nodes being added

        Parameters
        ----------
        input_schema : Schema
            Combined schema of the columns coming from upstream nodes
        selector : ColumnSelector
            Existing column selector for this node in the graph (often None)
        parents_selector : ColumnSelector
            Combined column selectors of parent nodes
        dependencies_selector : ColumnSelector
            Combined column selectors of dependency nodes

        Returns
        -------
        ColumnSelector
            Combined column selectors of parent and dependency nodes
        """
        self._validate_matching_cols(
            input_schema,
            parents_selector + dependencies_selector,
            self.compute_selector.__name__,
        )

        return parents_selector + dependencies_selector

    def compute_input_schema(
        self,
        root_schema: Schema,
        parents_schema: Schema,
        deps_schema: Schema,
        selector: ColumnSelector,
    ) -> Schema:
        """
        Combine schemas from the nodes being added

        Parameters
        ----------
        root_schema : Schema
            Schema of the columns from the input dataset
        parents_schema : Schema
            Schema of the columns from the parent nodes
        deps_schema : Schema
            Schema of the columns from the dependency nodes
        selector : ColumnSelector
            Existing column selector for this node in the graph (often None)

        Returns
        -------
        Schema
            Combined schema of columns from parents and dependencies
        """
        return parents_schema + deps_schema

    def transform(self, col_selector: ColumnSelector, df: DataFrameType) -> DataFrameType:
        """Simply returns the selected output columns from the input dataframe

        The main functionality of this operator has to do with computing the schemas
        for `+` nodes in the Workflow graph, so very little has to happen in the
        `transform` method.

        Parameters
        -----------
        columns: list of str or list of list of str
            The columns to apply this operator to
        df: Dataframe
            A pandas or cudf dataframe that this operator will work on

        Returns
        -------
        DataFrame
            Returns a transformed dataframe for this operator
        """
        return super()._get_columns(df, col_selector)

    @property
    def label(self) -> str:
        """
        Display name of this operator
        """
        return self._label



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from merlin.dag import ColumnSelector
from merlin.dag.ops.selection import SelectionOp
from merlin.schema import ColumnSchema, Schema


@pytest.mark.parametrize("engine", ["parquet"])
def test_selection_transform(df):
    selector = ColumnSelector(["x", "y"])
    op = SelectionOp(selector)

    result_df = op.transform(ColumnSelector(), df)

    assert (result_df.columns == ["x", "y"]).all()


@pytest.mark.parametrize("engine", ["parquet"])
def test_selection_output_column_names(df):
    selector = ColumnSelector(["x", "y"])

    op = SelectionOp(selector)
    result_selector = op.output_column_names(selector)

    assert result_selector.names == ["x", "y"]


@pytest.mark.parametrize("engine", ["parquet"])
def test_selection_output_schema(df):
    selector = ColumnSelector(["x", "y"])
    schema = Schema([ColumnSchema(col) for col in df.columns])
    op = SelectionOp(selector)

    result_schema = op.compute_output_schema(schema, ColumnSelector())

    assert result_schema.column_names == ["x", "y"]



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import warnings

import cudf
import uavro as ua
from dask.base import tokenize
from dask.dataframe.core import new_dd_object

from .dataset_engine import DatasetEngine


class AvroDatasetEngine(DatasetEngine):
    """AvroDatasetEngine

    Uses `uavro` to decompose dataset into groups of avro blocks.
    Uses `cudf` to create new partitions.
    """

    def __init__(self, paths, part_size, storage_options=None, cpu=False, **kwargs):
        super().__init__(paths, part_size, storage_options=storage_options, cpu=cpu)
        if kwargs != {}:
            raise ValueError("Unexpected AvroDatasetEngine argument(s).")
        self.blocksize = part_size

        # Avro reader needs a list of files
        # (Assume flat directory structure if this is a dir)
        if len(self.paths) == 1 and self.fs.isdir(self.paths[0]):
            self.paths = self.fs.glob(self.fs.sep.join([self.paths[0], "*"]))

        if self.cpu:
            raise ValueError("cpu=True not supported for AvroDatasetEngine.")

    def to_ddf(self, columns=None, cpu=None):

        # Check if we are using cpu
        cpu = self.cpu if cpu is None else cpu
        if cpu:
            raise ValueError("cpu=True not supported for AvroDatasetEngine.")

        # Get list of pieces for each output
        pieces, meta = self.process_metadata(columns=columns)

        # TODO: Remove warning and avoid use of uavro in read_partition when
        # cudf#6529 is fixed (https://github.com/rapidsai/cudf/issues/6529)
        if len(pieces) > len(self.paths):
            warnings.warn(
                "Row-subset selection in cudf avro reader is currently broken. "
                "Using uavro engine until cudf#6529 is addressed. "
                "EXPECT POOR PERFORMANCE!! (compared to cuio-based reader)"
            )

        # Construct collection
        token = tokenize(self.fs, self.paths, self.part_size, columns)
        read_avro_name = "read-avro-partition-" + token
        dsk = {
            (read_avro_name, i): (AvroDatasetEngine.read_partition, self.fs, piece, columns)
            for i, piece in enumerate(pieces)
        }
        return new_dd_object(dsk, read_avro_name, meta.iloc[:0], [None] * (len(pieces) + 1))

    def to_cpu(self):
        raise ValueError("cpu=True not supported for AvroDatasetEngine.")

    def to_gpu(self):
        self.cpu = False

    def process_metadata(self, columns=None):

        with open(self.paths[0], "rb") as fo:
            header = ua.core.read_header(fo)

            # Use first block for metadata
            num_rows = header["blocks"][0]["nrows"]
            file_byte_count = header["blocks"][0]["size"]
            meta = cudf.io.read_avro(self.paths[0], skiprows=0, num_rows=num_rows)

            # Convert the desired in-memory GPU size to the expected
            # on-disk storage size (blocksize)
            df_byte_count = meta.memory_usage(deep=True).sum()
            self.blocksize = int(float(file_byte_count) / df_byte_count * self.part_size)

        # Break apart files at the "Avro block" granularity
        pieces = []
        for path in self.paths:
            file_size = self.fs.du(path)
            if file_size > self.blocksize:
                part_count = 0
                with open(path, "rb") as fo:
                    header = ua.core.read_header(fo)
                    ua.core.scan_blocks(fo, header, file_size)
                    blocks = header["blocks"]

                    file_row_offset, part_row_count = 0, 0
                    file_block_offset, part_block_count = 0, 0
                    file_byte_offset, part_byte_count = blocks[0]["offset"], 0

                    for i, block in enumerate(blocks):
                        part_row_count += block["nrows"]
                        part_block_count += 1
                        part_byte_count += block["size"]
                        if part_byte_count >= self.blocksize:
                            pieces.append(
                                {
                                    "path": path,
                                    "rows": (file_row_offset, part_row_count),
                                    "blocks": (file_block_offset, part_block_count),
                                    "bytes": (file_byte_offset, part_byte_count),
                                }
                            )
                            part_count += 1
                            file_row_offset += part_row_count
                            file_block_offset += part_block_count
                            file_byte_offset += part_byte_count
                            part_row_count = part_block_count = part_byte_count = 0

                    if part_block_count:
                        pieces.append(
                            {
                                "path": path,
                                "rows": (file_row_offset, part_row_count),
                                "blocks": (file_block_offset, part_block_count),
                                "bytes": (file_byte_offset, part_byte_count),
                            }
                        )
                        part_count += 1
                if part_count == 1:
                    # No need to specify a byte range since we
                    # will need to read the entire file anyway.
                    pieces[-1] = {"path": pieces[-1]["path"]}
            else:
                pieces.append({"path": path})

        return pieces, meta

    @classmethod
    def read_partition(cls, fs, piece, columns):

        path = piece["path"]
        if "rows" in piece:

            # See: (https://github.com/rapidsai/cudf/issues/6529)
            # Using `uavro` library for now. This means we must convert
            # data to pandas, and then to cudf (which is much slower
            # than `cudf.read_avro`). TODO: Once `num_rows` is fixed,
            # this can be changed to:
            #
            #   skiprows, num_rows = piece["rows"]
            #   df = cudf.io.read_avro(
            #       path, skiprows=skiprows, num_rows=num_rows
            #   )

            block_offset, part_blocks = piece["blocks"]
            file_size = fs.du(piece["path"])
            with fs.open(piece["path"], "rb") as fo:
                header = ua.core.read_header(fo)
                ua.core.scan_blocks(fo, header, file_size)
                header["blocks"] = header["blocks"][block_offset : block_offset + part_blocks]

                # Adjust the total row count
                nrows = 0
                for block in header["blocks"]:
                    nrows += block["nrows"]
                header["nrows"] = nrows

                # Read in as pandas and convert to cudf (avoid block scan)
                df = cudf.from_pandas(
                    ua.core.filelike_to_dataframe(fo, file_size, header, scan=False)
                )
        else:
            df = cudf.io.read_avro(path)

        # Deal with column selection
        if columns is None:
            columns = list(df.columns)
        return df[columns]



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import List, Union

import merlin
from merlin.schema import Tags


class ColumnSelector:
    """A ColumnSelector describes a group of columns to be transformed by Operators in a
    Graph. Operators can be applied to the selected columns by shifting (>>) operators
    on to the ColumnSelector, which returns a new Node with the transformations
    applied. This lets you define a graph of operations that makes up your Graph.

    Parameters
    ----------
    names: list of (str or tuple of str)
        The columns to select from the input Dataset. The elements of this list are strings
        indicating the column names in most cases, but can also be tuples of strings
        for feature crosses.
    subgroups, optional: list of ColumnSelector objects
        This provides an alternate syntax for grouping column names together (instead
        of nesting tuples inside the list of names)
    """

    def __init__(
        self,
        names: List[str] = None,
        subgroups: List["ColumnSelector"] = None,
        tags: List[Union[Tags, str]] = None,
    ):
        self._names = names if names is not None else []
        self._tags = tags if tags is not None else []
        self.subgroups = subgroups if subgroups is not None else []

        if isinstance(self._names, merlin.dag.Node):
            raise TypeError("ColumnSelectors can not contain Nodes")

        if isinstance(self._names, str):
            self._names = [self._names]

        if isinstance(self.subgroups, ColumnSelector):
            self.subgroups = [self.subgroups]

        plain_names = []
        for name in self._names:
            if isinstance(name, str):
                plain_names.append(name)
            elif isinstance(name, merlin.dag.Node):
                raise ValueError("ColumnSelectors can not contain Nodes")
            elif isinstance(name, ColumnSelector):
                self.subgroups.append(name)
            else:
                self.subgroups.append(ColumnSelector(name))
        self._names = plain_names
        self._nested_check()

    @property
    def tags(self):
        return list(dict.fromkeys(self._tags).keys())

    @property
    def names(self):
        names = []
        names += self._names
        for subgroup in self.subgroups:
            names += subgroup.names

        # Only return unique column names
        return list(dict.fromkeys(names).keys())

    @property
    def grouped_names(self):
        names = []
        names += self._names
        for subgroup in self.subgroups:
            names.append(tuple(subgroup.names))

        # Only return unique grouped column names
        return list(dict.fromkeys(names).keys())

    def _nested_check(self, nests=0):
        if nests > 1:
            raise AttributeError("Too many nested subgroups")
        for col_sel0 in self.subgroups:
            col_sel0._nested_check(nests=nests + 1)

    def __add__(self, other):
        if other is None:
            return self
        elif isinstance(other, merlin.dag.Node):
            return other + self
        elif isinstance(other, ColumnSelector):

            return ColumnSelector(
                self._names + other._names,
                self.subgroups + other.subgroups,
                tags=self._tags + other._tags,
            )
        elif isinstance(other, Tags):
            return ColumnSelector(self._names, self.subgroups, tags=self._tags + [other])
        else:
            if isinstance(other, str):
                other = [other]
            return ColumnSelector(self._names + other, self.subgroups)

    def __radd__(self, other):
        return self + other

    def __rshift__(self, operator):
        if isinstance(operator, type) and issubclass(operator, merlin.dag.BaseOperator):
            # handle case where an operator class is passed
            operator = operator()

        return operator.create_node(self) >> operator

    def __eq__(self, other):
        if not isinstance(other, ColumnSelector):
            return False
        return other._names == self._names and other.subgroups == self.subgroups

    def __bool__(self):
        return bool(self._names or self.subgroups or self.tags)

    def resolve(self, schema):
        """Takes a schema and produces a new selector with selected column names
        how selection occurs (tags, name) does not matter."""
        # get names from tags or names
        root_selector = ColumnSelector(names=self._names, tags=self.tags)
        new_schema = schema.apply(root_selector)
        new_selector = ColumnSelector(new_schema.column_names)
        for group in self.subgroups:
            new_selector.subgroups.append(group.resolve(schema))
        return new_selector

    def filter_columns(self, other_selector):
        remaining_names = []
        remaining_groups = []

        for col in self._names:
            if col not in other_selector._names:
                remaining_names.append(col)

        for group in self.subgroups:
            if group not in other_selector.subgroups and all(
                col not in other_selector._names for col in group.names
            ):
                remaining_groups.append(group)

        return ColumnSelector(remaining_names, subgroups=remaining_groups)



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from merlin.schema.tags import Tags, TagSet


def test_tagset_init_normalizes_tags_to_enum():
    origin_tags = ["continuous", "list", "custom_tag"]
    tag_set = TagSet(origin_tags)
    assert Tags.CONTINUOUS in tag_set._tags
    assert Tags.LIST in tag_set._tags


def test_tagset_init_collision_error():
    with pytest.raises(ValueError) as err:
        tag_set = TagSet(["continuous", "categorical"])  # noqa

    assert "continuous" in str(err.value)
    assert "categorical" in str(err.value)
    assert "incompatible" in str(err.value)


def test_tagset_is_iterable():
    origin_tags = ["continuous", "list"]
    tag_set = TagSet(origin_tags)
    for tag in tag_set:
        assert tag.value in origin_tags
    assert len(tag_set) == len(origin_tags)


def test_tagset_add():
    origin_tags = [Tags.CONTINUOUS, Tags.LIST, "custom_tag"]
    tag_set = TagSet(origin_tags)

    new_tags = "custom_tag2"
    new_tag_set = tag_set + new_tags
    assert len(new_tag_set) == 4
    assert new_tags in new_tag_set
    assert all(origin_tag in new_tag_set for origin_tag in origin_tags)

    new_tags = ["custom_tag2"]
    new_tag_set = tag_set + new_tags

    assert len(new_tag_set) == 4
    assert all(new_tag in new_tag_set for new_tag in new_tags)
    assert all(origin_tag in new_tag_set for origin_tag in origin_tags)

    new_tags = TagSet(["custom_tag2"])
    new_tag_set = tag_set + new_tags

    assert len(new_tag_set) == 4
    assert all(tag in new_tag_set for tag in new_tags._tags)
    assert all(origin_tag in new_tag_set for origin_tag in origin_tags)


def test_tagset_sub():
    origin_tags = [Tags.CONTINUOUS, Tags.LIST]
    tag_set = TagSet(origin_tags + ["custom_tag"])
    assert len(tag_set) == 3

    new_tags = "custom_tag"
    new_tag_set = tag_set - new_tags
    assert len(new_tag_set) == 2
    assert new_tags not in new_tag_set
    assert all(origin_tag in new_tag_set for origin_tag in origin_tags)

    new_tags = ["custom_tag"]
    new_tag_set = tag_set - new_tags

    assert len(new_tag_set) == 2
    assert all(new_tag not in new_tag_set for new_tag in new_tags)
    assert all(origin_tag in new_tag_set for origin_tag in origin_tags)

    new_tags = TagSet(["custom_tag"])
    new_tag_set = tag_set - new_tags

    assert len(new_tag_set) == 2
    assert all(new_tag not in new_tag_set for new_tag in new_tags)
    assert all(origin_tag in new_tag_set for origin_tag in origin_tags)


def test_tagset_add_collision_error():
    origin_tags = ["continuous", "list", "custom_tag"]
    tag_set = TagSet(origin_tags)

    new_tags = "categorical"

    with pytest.raises(ValueError) as err:
        tag_set + new_tags  # pylint: disable=W0104

    assert "continuous" in str(err.value)
    assert "categorical" in str(err.value)
    assert "incompatible" in str(err.value)

    new_tags = ["categorical"]

    with pytest.raises(ValueError) as err:
        tag_set + new_tags  # pylint: disable=W0104

    assert "continuous" in str(err.value)
    assert "categorical" in str(err.value)
    assert "incompatible" in str(err.value)

    new_tags = TagSet(["categorical"])

    with pytest.raises(ValueError) as err:
        tag_set + new_tags  # pylint: disable=W0104

    assert "continuous" in str(err.value)
    assert "categorical" in str(err.value)
    assert "incompatible" in str(err.value)
